
#
# spyne - Copyright (C) Spyne contributors.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
#

"""
This module contains the :class:`Service` class and its helper objects.
"""

import logging
logger = logging.getLogger(__name__)

from spyne.util.six.moves.collections_abc import Sequence

from spyne.evmgr import EventManager
from spyne.util import six
from spyne.util.oset import oset


class ServiceBaseMeta(type):
    """Adds event managers."""

    def __init__(self, cls_name, cls_bases, cls_dict):
        super(ServiceBaseMeta, self).__init__(cls_name, cls_bases, cls_dict)

        self.public_methods = {}
        self.event_manager = EventManager(self,
                                      self.__get_base_event_handlers(cls_bases))

    def __get_base_event_handlers(self, cls_bases):
        handlers = {}

        for base in cls_bases:
            evmgr = getattr(base, 'event_manager', None)
            if evmgr is None:
                continue

            for k, v in evmgr.handlers.items():
                handler = handlers.get(k, oset())
                for h in v:
                    handler.add(h)
                handlers[k] = handler

        return handlers

class ServiceMeta(ServiceBaseMeta):
    """Creates the :class:`spyne.MethodDescriptor` objects by iterating over
    tagged methods.
    """

    def __init__(self, cls_name, cls_bases, cls_dict):
        super(ServiceMeta, self).__init__(cls_name, cls_bases, cls_dict)

        self.__has_aux_methods = self.__aux__ is not None
        has_nonaux_methods = None

        for k, v in cls_dict.items():
            if not hasattr(v, '_is_rpc'):
                continue

            descriptor = v(_default_function_name=k, _service_class=self)

            # these two lines are needed for staticmethod wrapping to work
            setattr(self, k, staticmethod(descriptor.function))
            descriptor.reset_function(getattr(self, k))

            try:
                getattr(self, k).descriptor = descriptor
            except AttributeError:
                pass
                # FIXME: this fails with builtins. Temporary hack while we
                # investigate whether we really need this or not

            self.public_methods[k] = descriptor
            if descriptor.aux is None and self.__aux__ is None:
                has_nonaux_methods = True
            else:
                self.__has_aux_methods = True

            if self.__has_aux_methods and has_nonaux_methods:
                raise Exception("You can't mix primary and "
                                "auxiliary methods in a single service definition.")

    def is_auxiliary(self):
        return self.__has_aux_methods


# FIXME: To be renamed to ServiceBase in Spyne 3
@six.add_metaclass(ServiceBaseMeta)
class ServiceBaseBase(object):
    __in_header__ = None
    """The incoming header object that the methods under this service definition
    accept."""

    __out_header__ = None
    """The outgoing header object that the methods under this service definition
    accept."""

    __service_name__ = None
    """The name of this service definition as exposed in the interface document.
    Defaults to the class name."""

    __service_module__ = None
    """This is used for internal idenfitication of the service class,
    to override the ``__module__`` attribute."""

    __port_types__ = ()
    """WSDL-Specific portType mappings"""

    __aux__ = None
    """The auxiliary method type. When set, the ``aux`` property of every method
    defined under this service is set to this value. The _aux flag in the @srpc
    decorator overrides this."""

    @classmethod
    def get_service_class_name(cls):
        return cls.__name__

    @classmethod
    def get_service_name(cls):
        if cls.__service_name__ is None:
            return cls.__name__
        else:
            return cls.__service_name__

    @classmethod
    def get_service_module(cls):
        if cls.__service_module__ is None:
            return cls.__module__
        else:
            return cls.__service_module__

    @classmethod
    def get_internal_key(cls):
        return "%s.%s" % (cls.get_service_module(), cls.get_service_name())

    @classmethod
    def get_port_types(cls):
        return cls.__port_types__

    @classmethod
    def _has_callbacks(cls):
        """Determines if this service definition has callback methods or not."""

        for method in cls.public_methods.values():
            if method.is_callback:
                return True

        return False

    @classmethod
    def get_context(cls):
        """Returns a user defined context. Override this in your ServiceBase
        subclass to customize context generation."""
        return None

    @classmethod
    def call_wrapper(cls, ctx, args=None):
        """Called in place of the original method call. You can override this to
        do your own exception handling.

        :param ctx: The method context.

        The overriding function must call this function by convention.
        """

        if ctx.function is not None:
            if args is None:
                args = ctx.in_object

                assert not isinstance(args, six.string_types)

                # python3 wants a proper sequence as *args
                if not isinstance(args, Sequence):
                    args = tuple(args)

                if not ctx.descriptor.no_ctx:
                    args = (ctx,) + tuple(args)

            return ctx.function(*args)

    @classmethod
    def initialize(cls, app):
        pass


@six.add_metaclass(ServiceMeta)
class Service(ServiceBaseBase):
    """The ``Service`` class is the base class for all service definitions.

    The convention is to have public methods defined under a subclass of this
    class along with common properties of public methods like header classes or
    auxiliary processors. The :func:`spyne.decorator.srpc` decorator or its
    wrappers should be used to flag public methods.

    This class is designed to be subclassed just once. You're supposed to
    combine Service subclasses in order to get the public method mix you
    want.

    It is a natural abstract base class, because it's of no use without any
    method definitions, hence the 'Base' suffix in the name.

    This class supports the following events:
        * ``method_call``
            Called right before the service method is executed

        * ``method_return_object``
            Called right after the service method is executed

        * ``method_exception_object``
            Called when an exception occurred in a service method, before the
            exception is serialized.

        * ``method_accept_document``
            Called by the transport right after the incoming stream is parsed to
            the incoming protocol's document type.

        * ``method_return_document``
            Called by the transport right after the outgoing object is
            serialized to the outgoing protocol's document type.

        * ``method_exception_document``
            Called by the transport right before the outgoing exception object
            is serialized to the outgoing protocol's document type.

        * ``method_return_string``
            Called by the transport right before passing the return string to
            the client.

        * ``method_exception_string``
            Called by the transport right before passing the exception string to
            the client.
    """


# FIXME: To be deleted in Spyne 3
ServiceBase = Service



#!/usr/bin/env python
#
# spyne - Copyright (C) Spyne contributors.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
#

import logging
suds_logger = logging.getLogger('suds')
suds_logger.setLevel(logging.INFO)

import unittest

from datetime import datetime
from base64 import b64encode, b64decode

from suds.sax.parser import Parser
from suds.client import Client
from suds.plugin import MessagePlugin
from suds import WebFault

from spyne.util import six

from spyne.test.interop._test_soap_client_base import SpyneClientTestBase, \
    server_started


class LastReceivedPlugin(MessagePlugin):
    def received(self, context):
        sax = Parser()
        self.reply = sax.parse(string=context.reply)


class TestSuds(SpyneClientTestBase, unittest.TestCase):
    def setUp(self):
        SpyneClientTestBase.setUp(self, 'http')

        port, = server_started.keys()

        self.client = Client("http://localhost:%d/?wsdl" % port, cache=None,
                                                 plugins=[LastReceivedPlugin()])
        self.ns = "spyne.test.interop.server"

    def test_echo_datetime(self):
        val = datetime.now()
        ret = self.client.service.echo_datetime(val)

        assert val == ret

    def test_echo_datetime_with_invalid_format(self):
        val = datetime.now()
        ret = self.client.service.echo_datetime_with_invalid_format(val)

        assert val == ret

    def test_echo_date(self):
        val = datetime.now().date()
        ret = self.client.service.echo_date(val)

        assert val == ret

    def test_echo_date_with_invalid_format(self):
        val = datetime.now().date()
        ret = self.client.service.echo_date_with_invalid_format(val)

        assert val == ret

    def test_echo_time(self):
        val = datetime.now().time()
        ret = self.client.service.echo_time(val)

        assert val == ret

    def test_echo_time_with_invalid_format(self):
        val = datetime.now().time()
        ret = self.client.service.echo_time_with_invalid_format(val)

        assert val == ret

    def test_echo_simple_boolean_array(self):
        val = [False, False, False, True]
        ret = self.client.service.echo_simple_boolean_array(val)

        assert val == ret

    def test_echo_boolean(self):
        val = True
        ret = self.client.service.echo_boolean(val)
        self.assertEqual(val, ret)

        val = False
        ret = self.client.service.echo_boolean(val)
        self.assertEqual(val, ret)

    def test_enum(self):
        DaysOfWeekEnum = self.client.factory.create("DaysOfWeekEnum")

        val = DaysOfWeekEnum.Monday
        ret = self.client.service.echo_enum(val)

        assert val == ret

    def test_bytearray(self):
        val = b"\x00\x01\x02\x03\x04"
        # suds doesn't support base64 encoding, so we do it manually
        ret = self.client.service.echo_bytearray(b64encode(val).decode())

        assert val == b64decode(ret)

    def test_validation(self):
        non_nillable_class = self.client.factory.create(
                                                  "{hunk.sunk}NonNillableClass")
        non_nillable_class.i = 6
        non_nillable_class.s = None

        try:
            self.client.service.non_nillable(non_nillable_class)
        except WebFault as e:
            pass
        else:
            raise Exception("must fail")

    def test_echo_integer_array(self):
        ia = self.client.factory.create('integerArray')
        ia.integer.extend([1, 2, 3, 4, 5])
        self.client.service.echo_integer_array(ia)

    def test_echo_in_header(self):
        in_header = self.client.factory.create('InHeader')
        in_header.s = 'a'
        in_header.i = 3

        self.client.set_options(soapheaders=in_header)
        ret = self.client.service.echo_in_header()
        self.client.set_options(soapheaders=None)

        print(ret)

        self.assertEqual(in_header.s, ret.s)
        self.assertEqual(in_header.i, ret.i)

    def test_echo_in_complex_header(self):
        in_header = self.client.factory.create('InHeader')
        in_header.s = 'a'
        in_header.i = 3
        in_trace_header = self.client.factory.create('InTraceHeader')
        in_trace_header.client = 'suds'
        in_trace_header.callDate = datetime(year=2000, month=1, day=1, hour=0,
                                              minute=0, second=0, microsecond=0)

        self.client.set_options(soapheaders=(in_header, in_trace_header))
        ret = self.client.service.echo_in_complex_header()
        self.client.set_options(soapheaders=None)

        print(ret)

        self.assertEqual(in_header.s, ret[0].s)
        self.assertEqual(in_header.i, ret[0].i)
        self.assertEqual(in_trace_header.client, ret[1].client)
        self.assertEqual(in_trace_header.callDate, ret[1].callDate)

    def test_send_out_header(self):
        out_header = self.client.factory.create('OutHeader')
        out_header.dt = datetime(year=2000, month=1, day=1)
        out_header.f = 3.141592653

        ret = self.client.service.send_out_header()

        self.assertTrue(isinstance(ret, type(out_header)))
        self.assertEqual(ret.dt, out_header.dt)
        self.assertEqual(ret.f, out_header.f)

    def test_send_out_complex_header(self):
        out_header = self.client.factory.create('OutHeader')
        out_header.dt = datetime(year=2000, month=1, day=1)
        out_header.f = 3.141592653
        out_trace_header = self.client.factory.create('OutTraceHeader')
        out_trace_header.receiptDate = datetime(year=2000, month=1, day=1,
                                  hour=1, minute=1, second=1, microsecond=1)
        out_trace_header.returnDate = datetime(year=2000, month=1, day=1,
                                 hour=1, minute=1, second=1, microsecond=100)

        ret = self.client.service.send_out_complex_header()

        self.assertTrue(isinstance(ret[0], type(out_header)))
        self.assertEqual(ret[0].dt, out_header.dt)
        self.assertEqual(ret[0].f, out_header.f)
        self.assertTrue(isinstance(ret[1], type(out_trace_header)))
        self.assertEqual(ret[1].receiptDate, out_trace_header.receiptDate)
        self.assertEqual(ret[1].returnDate, out_trace_header.returnDate)
        # Control the reply soap header (in an unelegant way but this is the
        # only way with suds)
        soapheaders = self.client.options.plugins[0].reply.getChild("Envelope").getChild("Header")
        soap_out_header = soapheaders.getChild('OutHeader')
        self.assertEqual('T'.join((out_header.dt.date().isoformat(),
                                    out_header.dt.time().isoformat())),
                          soap_out_header.getChild('dt').getText())
        self.assertEqual(six.text_type(out_header.f), soap_out_header.getChild('f').getText())
        soap_out_trace_header = soapheaders.getChild('OutTraceHeader')
        self.assertEqual('T'.join((out_trace_header.receiptDate.date().isoformat(),
                                    out_trace_header.receiptDate.time().isoformat())),
                          soap_out_trace_header.getChild('receiptDate').getText())
        self.assertEqual('T'.join((out_trace_header.returnDate.date().isoformat(),
                                    out_trace_header.returnDate.time().isoformat())),
                          soap_out_trace_header.getChild('returnDate').getText())

    def test_echo_string(self):
        test_string = "OK"
        ret = self.client.service.echo_string(test_string)

        self.assertEqual(ret, test_string)

    def __get_xml_test_val(self):
        return {
            "test_sub": {
                "test_subsub1": {
                    "test_subsubsub1": ["subsubsub1 value"]
                },
                "test_subsub2": ["subsub2 value 1", "subsub2 value 2"],
                "test_subsub3": [
                    {
                        "test_subsub3sub1": ["subsub3sub1 value"]
                    },
                    {
                        "test_subsub3sub2": ["subsub3sub2 value"]
                    },
                ],
                "test_subsub4": [],
                "test_subsub5": ["x"],
            }
        }


    def test_echo_simple_class(self):
        val = self.client.factory.create("{spyne.test.interop.server}SimpleClass")

        val.i = 45
        val.s = "asd"

        ret = self.client.service.echo_simple_class(val)

        assert ret.i == val.i
        assert ret.s == val.s

    def test_echo_class_with_self_reference(self):
        val = self.client.factory.create("{spyne.test.interop.server}ClassWithSelfReference")

        val.i = 45
        val.sr = self.client.factory.create("{spyne.test.interop.server}ClassWithSelfReference")
        val.sr.i = 50
        val.sr.sr = None

        ret = self.client.service.echo_class_with_self_reference(val)

        assert ret.i == val.i
        assert ret.sr.i == val.sr.i

    def test_echo_nested_class(self):
        val = self.client.factory.create("{punk.tunk}NestedClass");

        val.i = 45
        val.s = "asd"
        val.f = 12.34
        val.ai = self.client.factory.create("integerArray")
        val.ai.integer.extend([1, 2, 3, 45, 5, 3, 2, 1, 4])

        val.simple = self.client.factory.create("{spyne.test.interop.server}SimpleClassArray")

        val.simple.SimpleClass.append(self.client.factory.create("{spyne.test.interop.server}SimpleClass"))
        val.simple.SimpleClass.append(self.client.factory.create("{spyne.test.interop.server}SimpleClass"))

        val.simple.SimpleClass[0].i = 45
        val.simple.SimpleClass[0].s = "asd"
        val.simple.SimpleClass[1].i = 12
        val.simple.SimpleClass[1].s = "qwe"

        val.other = self.client.factory.create("{spyne.test.interop.server}OtherClass");
        val.other.dt = datetime.now()
        val.other.d = 123.456
        val.other.b = True

        ret = self.client.service.echo_nested_class(val)

        self.assertEqual(ret.i, val.i)
        self.assertEqual(ret.ai[0], val.ai[0])
        self.assertEqual(ret.simple.SimpleClass[0].s, val.simple.SimpleClass[0].s)
        self.assertEqual(ret.other.dt, val.other.dt)

    def test_huge_number(self):
        self.assertEqual(self.client.service.huge_number(), 2 ** int(1e5))

    def test_long_string(self):
        self.assertEqual(self.client.service.long_string(),
                                                   ('0123456789abcdef' * 16384))

    def test_empty(self):
        self.client.service.test_empty()

    def test_echo_extension_class(self):
        val = self.client.factory.create("{bar}ExtensionClass")

        val.i = 45
        val.s = "asd"
        val.f = 12.34

        val.simple = self.client.factory.create("{spyne.test.interop.server}SimpleClassArray")

        val.simple.SimpleClass.append(self.client.factory.create("{spyne.test.interop.server}SimpleClass"))
        val.simple.SimpleClass.append(self.client.factory.create("{spyne.test.interop.server}SimpleClass"))

        val.simple.SimpleClass[0].i = 45
        val.simple.SimpleClass[0].s = "asd"
        val.simple.SimpleClass[1].i = 12
        val.simple.SimpleClass[1].s = "qwe"

        val.other = self.client.factory.create("{spyne.test.interop.server}OtherClass");
        val.other.dt = datetime.now()
        val.other.d = 123.456
        val.other.b = True

        val.p = self.client.factory.create("{hunk.sunk}NonNillableClass");
        val.p.dt = datetime(2010, 6, 2)
        val.p.i = 123
        val.p.s = "punk"

        val.l = datetime(2010, 7, 2)
        val.q = 5

        ret = self.client.service.echo_extension_class(val)
        print(ret)

        self.assertEqual(ret.i, val.i)
        self.assertEqual(ret.s, val.s)
        self.assertEqual(ret.f, val.f)
        self.assertEqual(ret.simple.SimpleClass[0].i, val.simple.SimpleClass[0].i)
        self.assertEqual(ret.other.dt, val.other.dt)
        self.assertEqual(ret.p.s, val.p.s)


    def test_python_exception(self):
        try:
            self.client.service.python_exception()
            raise Exception("must fail")
        except WebFault as e:
            pass

    def test_soap_exception(self):
        try:
            self.client.service.soap_exception()
            raise Exception("must fail")
        except WebFault as e:
            pass

    def test_complex_return(self):
        ret = self.client.service.complex_return()

        self.assertEqual(ret.resultCode, 1)
        self.assertEqual(ret.resultDescription, "Test")
        self.assertEqual(ret.transactionId, 123)
        self.assertEqual(ret.roles.RoleEnum[0], "MEMBER")

    def test_return_invalid_data(self):
        try:
            self.client.service.return_invalid_data()
            raise Exception("must fail")
        except:
            pass

    def test_custom_messages(self):
        ret = self.client.service.custom_messages("test")

        assert ret == 'test'

    def test_echo_simple_bare(self):
        ret = self.client.service.echo_simple_bare("test")

        assert ret == 'test'

    #
    # This test is disabled because suds does not create the right request
    # object. Opening the first <ns0:string> tag below is wrong.
    #
    #<SOAP-ENV:Envelope xmlns:ns0="spyne.test.interop.server"
    #                   xmlns:xs="http://www.w3.org/2001/XMLSchema"
    #                   xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
    #                   xmlns:ns1="http://schemas.xmlsoap.org/soap/envelope/"
    #                   xmlns:SOAP-ENV="http://schemas.xmlsoap.org/soap/envelope/">
    #  <SOAP-ENV:Header/>
    #  <ns1:Body>
    #      <ns0:echo_complex_bare>
    #         <ns0:string>
    #            <ns0:string>abc</ns0:string>
    #            <ns0:string>def</ns0:string>
    #         </ns0:string>
    #      </ns0:echo_complex_bare>
    #  </ns1:Body>
    #</SOAP-ENV:Envelope>
    #
    # The right request looks like this:
    #
    #      <ns0:echo_complex_bare>
    #         <ns0:string>abc</ns0:string>
    #         <ns0:string>def</ns0:string>
    #      </ns0:echo_complex_bare>
    #
    def _test_echo_complex_bare(self):
        val = ['abc','def']
        ia = self.client.factory.create('stringArray')
        ia.string.extend(val)
        ret = self.client.service.echo_complex_bare(ia)

        assert ret == val


if __name__ == '__main__':
    unittest.main()




#
# spyne - Copyright (C) Spyne contributors.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
#

from __future__ import absolute_import

import re
import spyne
import datetime

from spyne.model import SimpleModel
from spyne.model.primitive import NATIVE_MAP

FLOAT_PATTERN = r'-?[0-9]+\.?[0-9]*(e-?[0-9]+)?'
DATE_PATTERN = r'(?P<year>\d{4})-(?P<month>\d{2})-(?P<day>\d{2})'
TIME_PATTERN = r'(?P<hr>\d{2}):(?P<min>\d{2}):(?P<sec>\d{2})(?P<sec_frac>\.\d+)?'
OFFSET_PATTERN = r'(?P<tz_hr>[+-]\d{2}):(?P<tz_min>\d{2})'
DATETIME_PATTERN = DATE_PATTERN + '[T ]' + TIME_PATTERN


class Time(SimpleModel):
    """Just that, Time. No time zone support.

    Native type is :class:`datetime.time`.
    """

    __type_name__ = 'time'
    Value = datetime.time

    class Attributes(SimpleModel.Attributes):
        """Customizable attributes of the :class:`spyne.model.primitive.Time`
        type."""

        gt = None  # minExclusive
        """The time should be greater than this time."""

        ge = datetime.time(0, 0, 0, 0)  # minInclusive
        """The time should be greater than or equal to this time."""

        lt = None  # maxExclusive
        """The time should be lower than this time."""

        le = datetime.time(23, 59, 59, 999999)  # maxInclusive
        """The time should be lower than or equal to this time."""

        pattern = None
        """A regular expression that matches the whole time. See here for more
        info: http://www.regular-expressions.info/xml.html"""

        time_format = None
        """Time format fed to the ``strftime`` function. See:
        http://docs.python.org/library/datetime.html?highlight=strftime#strftime-strptime-behavior
        Ignored by protocols like SOAP which have their own ideas about how
        Date objects should be serialized."""

    @staticmethod
    def is_default(cls):
        return (    SimpleModel.is_default(cls)
                and cls.Attributes.gt == Time.Attributes.gt
                and cls.Attributes.ge == Time.Attributes.ge
                and cls.Attributes.lt == Time.Attributes.lt
                and cls.Attributes.le == Time.Attributes.le
                and cls.Attributes.pattern == Time.Attributes.pattern
        )

    @staticmethod
    def validate_native(cls, value):
        return SimpleModel.validate_native(cls, value) and (
            value is None or (
                (cls.Attributes.gt is None or value >  cls.Attributes.gt)
                and value >= cls.Attributes.ge
                and (cls.Attributes.lt is None or value <  cls.Attributes.lt)
                and value <= cls.Attributes.le
            ))

_min_dt = datetime.datetime.min.replace(tzinfo=spyne.LOCAL_TZ)
_max_dt = datetime.datetime.max.replace(tzinfo=spyne.LOCAL_TZ)


class DateTime(SimpleModel):
    """A compact way to represent dates and times together. Supports time zones.
    Working with timezones is a bit quirky -- Spyne works very hard to have
    all datetimes with time zones internally and only strips them when
    explicitly requested with ``timezone=False``\\. See
    :attr:`DateTime.Attributes.as_timezone` for more information.

    Native type is :class:`datetime.datetime`.
    """

    __type_name__ = 'dateTime'
    Value = datetime.datetime

    _local_re = re.compile(DATETIME_PATTERN)
    _utc_re = re.compile(DATETIME_PATTERN + 'Z')
    _offset_re = re.compile(DATETIME_PATTERN + OFFSET_PATTERN)

    class Attributes(SimpleModel.Attributes):
        """Customizable attributes of the :class:`spyne.model.primitive.DateTime`
        type."""

        gt = None  # minExclusive
        """The datetime should be greater than this datetime. It must always
        have a timezone."""

        ge = _min_dt  # minInclusive
        """The datetime should be greater than or equal to this datetime. It
        must always have a timezone."""

        lt = None  # maxExclusive
        """The datetime should be lower than this datetime. It must always have
        a timezone."""

        le = _max_dt  # maxInclusive
        """The datetime should be lower than or equal to this datetime. It must
        always have a timezone."""

        pattern = None
        """A regular expression that matches the whole datetime. See here for
        more info: http://www.regular-expressions.info/xml.html"""

        dt_format = None
        """DateTime format fed to the ``strftime`` function. See:
        http://docs.python.org/library/datetime.html?highlight=strftime#strftime-strptime-behavior
        Ignored by protocols like SOAP which have their own ideas about how
        DateTime objects should be serialized."""

        out_format = None
        """DateTime format fed to the ``strftime`` function only when
        serializing. See:
        http://docs.python.org/library/datetime.html?highlight=strftime#strftime-strptime-behavior
        Ignored by protocols like SOAP which have their own ideas about how
        DateTime objects should be serialized."""

        string_format = None
        """A regular python string formatting string. %s will contain the date
        string. See here for more info:
        http://docs.python.org/library/stdtypes.html#string-formatting"""

        as_timezone = None
        """When not None, converts:
            - Outgoing values to the given time zone (by calling
              ``.astimezone()``).
            - Incoming values without tzinfo to the given time zone by calling
              ``.replace(tzinfo=<as_timezone>)`` and values with tzinfo to the
               given timezone by calling ``.astimezone()``.

        Either None or a return value of pytz.timezone()

        When this is None and a datetime with tzinfo=None comes in, it's
        converted to spyne.LOCAL_TZ which defaults to ``pytz.utc``. You can use
        `tzlocal <https://pypi.python.org/pypi/tzlocal>`_ to set it to local
        time right after ``import spyne``.
        """

        timezone = True
        """If False, time zone info is stripped before serialization. Also makes
        sqlalchemy schema generator emit 'timestamp without timezone'."""

        serialize_as = None
        """One of (None, 'sec', 'sec_float', 'msec', 'msec_float', 'usec')"""

        # TODO: Move this to ModelBase and make it work with all types in all
        # protocols.
        parser = None
        """Callable for string parser. It must accept exactly four arguments:
        `protocol, cls, string` and must return a `datetime.datetime` object.
        If this is not None, all other parsing configurations (e.g.
        `date_format`) are ignored.
        """

    @staticmethod
    def is_default(cls):
        return (    SimpleModel.is_default(cls)
                and cls.Attributes.gt == DateTime.Attributes.gt
                and cls.Attributes.ge == DateTime.Attributes.ge
                and cls.Attributes.lt == DateTime.Attributes.lt
                and cls.Attributes.le == DateTime.Attributes.le
                and cls.Attributes.pattern == DateTime.Attributes.pattern
        )

    @staticmethod
    def validate_native(cls, value):
        if isinstance(value, datetime.datetime) and value.tzinfo is None:
            value = value.replace(tzinfo=spyne.LOCAL_TZ)
        return SimpleModel.validate_native(cls, value) and (
            value is None or (
                # min_dt is also a valid value if gt is intact.
                    (cls.Attributes.gt is None or value > cls.Attributes.gt)
                and value >= cls.Attributes.ge
                # max_dt is also a valid value if lt is intact.
                and (cls.Attributes.lt is None or value < cls.Attributes.lt)
                and value <= cls.Attributes.le
            ))


class Date(DateTime):
    """Just that, Date. No time zone support.

    Native type is :class:`datetime.date`.
    """

    __type_name__ = 'date'

    _offset_re = re.compile(DATE_PATTERN + '(' + OFFSET_PATTERN + '|Z)')
    Value = datetime.date

    class Attributes(DateTime.Attributes):
        """Customizable attributes of the :class:`spyne.model.primitive.Date`
        type."""

        gt = None  # minExclusive
        """The date should be greater than this date."""

        ge = datetime.date(1, 1, 1)  # minInclusive
        """The date should be greater than or equal to this date."""

        lt = None  # maxExclusive
        """The date should be lower than this date."""

        le = datetime.date(datetime.MAXYEAR, 12, 31)  # maxInclusive
        """The date should be lower than or equal to this date."""

        date_format = None
        """Date format fed to the ``strftime`` function. See:
        http://docs.python.org/library/datetime.html?highlight=strftime#strftime-strptime-behavior
        Ignored by protocols like SOAP which have their own ideas about how
        Date objects should be serialized."""

        pattern = None
        """A regular expression that matches the whole date. See here for more
        info: http://www.regular-expressions.info/xml.html"""


    @staticmethod
    def is_default(cls):
        return (    SimpleModel.is_default(cls)
                and cls.Attributes.gt == Date.Attributes.gt
                and cls.Attributes.ge == Date.Attributes.ge
                and cls.Attributes.lt == Date.Attributes.lt
                and cls.Attributes.le == Date.Attributes.le
                and cls.Attributes.pattern == Date.Attributes.pattern
        )


# this object tries to follow ISO 8601 standard.
class Duration(SimpleModel):
    """Native type is :class:`datetime.timedelta`."""

    __type_name__ = 'duration'
    Value = datetime.timedelta


NATIVE_MAP.update({
    datetime.datetime: DateTime,
    datetime.time: Time,
    datetime.date: Date,
    datetime.timedelta: Duration,
})



#!/usr/bin/env python
#
# spyne - Copyright (C) Spyne contributors.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301

import unittest

from webtest import TestApp as _TestApp  # avoid confusing py.test

from spyne.application import Application
from spyne.decorator import srpc
from spyne.service import Service
from spyne.model.primitive import Integer, Unicode
from spyne.model.complex import Iterable
from spyne.protocol.soap import Soap11
from spyne.protocol.http import HttpRpc
from spyne.protocol.json import JsonDocument
from spyne.server.wsgi import WsgiApplication

from spyne.const.xml import PREFMAP, NS_WSDL11_SOAP

def strip_whitespace(string):
    return ''.join(string.split())


class TestOperationRequestSuffix(unittest.TestCase):
    """
    test different protocols with REQUEST_SUFFIX and _operation_name
    _in_message_name is a concern, will test that as well
    """

    default_function_name = 'echo'

    # output is not affected, will use soap output for all tests
    result_body = '''
        <soap11env:Body>
            <tns:echoResponse>
                <tns:echoResult>
                    <tns:string>Echo, test</tns:string>
                    <tns:string>Echo, test</tns:string>
                </tns:echoResult>
            </tns:echoResponse>
        </soap11env:Body>'''

    def get_function_names(self, suffix, _operation_name=None,
                           _in_message_name=None):
        """This tests the logic of how names are produced.
        Its logic should match expected behavior of the decorator.
        returns operation name, in message name, service name depending on
        args"""
        function_name = self.default_function_name

        if _operation_name is None:
            operation_name = function_name
        else:
            operation_name = _operation_name

        if _in_message_name is None:
            request_name = operation_name + suffix
        else:
            request_name = _in_message_name

        return function_name, operation_name, request_name

    def get_app(self, in_protocol, suffix, _operation_name=None,
                _in_message_name=None):
        """setup testapp dependent on suffix and _in_message_name"""

        import spyne.const
        spyne.const.REQUEST_SUFFIX = suffix

        class EchoService(Service):

            srpc_kparams = {'_returns': Iterable(Unicode)}
            if _in_message_name:
                srpc_kparams['_in_message_name'] = _in_message_name
            if _operation_name:
                srpc_kparams['_operation_name'] = _operation_name

            @srpc(Unicode, Integer, **srpc_kparams)
            def echo(string, times):
                for i in range(times):
                    yield 'Echo, %s' % string

        application = Application([EchoService],
            tns='spyne.examples.echo',
            in_protocol=in_protocol,
            out_protocol=Soap11()
        )
        app = WsgiApplication(application)

        testapp = _TestApp(app)

        # so that it doesn't interfere with other tests.
        spyne.const.REQUEST_SUFFIX = ''

        return testapp

    def assert_response_ok(self, resp):
        """check the default response"""
        self.assertEqual(resp.status_int, 200, resp)
        self.assertTrue(
            strip_whitespace(self.result_body) in strip_whitespace(str(resp)),
            '{0} not in {1}'.format(self.result_body, resp))

    ### application error tests ###
    def assert_application_error(self, suffix, _operation_name=None,
                                                         _in_message_name=None):
        self.assertRaises(ValueError,
                          self.get_app, Soap11(validator='lxml'), suffix,
                          _operation_name, _in_message_name)

    def test_assert_application_error(self):
        """check error when op namd and in name are both used"""
        self.assert_application_error(suffix='',
                                      _operation_name='TestOperationName',
                                      _in_message_name='TestMessageName')

    ### soap tests ###
    def assert_soap_ok(self, suffix, _operation_name=None,
                                                         _in_message_name=None):
        """helper to test soap requests"""

        # setup
        app = self.get_app(Soap11(validator='lxml'), suffix, _operation_name,
                           _in_message_name)

        function_name, operation_name, request_name = self.get_function_names(
            suffix, _operation_name, _in_message_name)

        soap_input_body = """
        <SOAP-ENV:Envelope xmlns:ns0="spyne.examples.echo"
        xmlns:ns1="http://schemas.xmlsoap.org/soap/envelope/"
        xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
        xmlns:SOAP-ENV="http://schemas.xmlsoap.org/soap/envelope/">
           <SOAP-ENV:Header/>
           <ns1:Body>
              <ns0:{0}>
                 <ns0:string>test</ns0:string>
                 <ns0:times>2</ns0:times>
              </ns0:{0}>
           </ns1:Body>
        </SOAP-ENV:Envelope>""".format(request_name)

        # check wsdl
        wsdl = app.get('/?wsdl')
        self.assertEqual(wsdl.status_int, 200, wsdl)
        self.assertTrue(request_name in wsdl,
                        '{0} not found in wsdl'.format(request_name))

        soap_strings = [
            '<wsdl:operation name="{0}"'.format(operation_name),
            '<{0}:operation soapAction="{1}"'.format(PREFMAP[NS_WSDL11_SOAP], operation_name),
            '<wsdl:input name="{0}">'.format(request_name),
            '<xs:element name="{0}"'.format(request_name),
            '<xs:complexType name="{0}">'.format(request_name),
        ]
        for soap_string in soap_strings:
            self.assertTrue(soap_string in wsdl,
                            '{0} not in {1}'.format(soap_string, wsdl))
        if request_name != operation_name:
            wrong_string = '<wsdl:operation name="{0}"'.format(request_name)
            self.assertFalse(wrong_string in wsdl,
                             '{0} in {1}'.format(wrong_string, wsdl))

        output_name = '<wsdl:output name="{0}Response"'.format(
            self.default_function_name)
        self.assertTrue(output_name in wsdl,
                        'REQUEST_SUFFIX or _in_message_name changed the '
                        'output name, it should be: {0}'.format(
                            output_name))

        # check soap operation succeeded
        resp = app.post('/', soap_input_body,
                                    content_type='applicaion/xml; charset=utf8')
        self.assert_response_ok(resp)

    def test_soap_with_suffix(self):
        self.assert_soap_ok(suffix='Request')

    def test_soap_no_suffix(self):
        self.assert_soap_ok(suffix='')

    def test_soap_with_suffix_with_message_name(self):
        self.assert_soap_ok(suffix='Request',
                            _in_message_name='TestInMessageName')

    def test_soap_no_suffix_with_message_name(self):
        self.assert_soap_ok(suffix='', _in_message_name='TestInMessageName')

    def test_soap_with_suffix_with_operation_name(self):
        self.assert_soap_ok(suffix='Request',
                            _operation_name='TestOperationName')

    def test_soap_no_suffix_with_operation_name(self):
        self.assert_soap_ok(suffix='', _operation_name='TestOperationName')

    ### json tests ###
    def assert_json_ok(self, suffix, _operation_name=None,
                       _in_message_name=None):
        """helper to test json requests"""

        # setup
        app = self.get_app(JsonDocument(validator='soft'), suffix,
                           _operation_name, _in_message_name)

        function_name, operation_name, request_name = self.get_function_names(
            suffix, _operation_name, _in_message_name)

        json_input_body = '{"' + request_name + '": {"string": "test", ' \
                                                '"times": 2}}'

        # check json operation succeeded
        resp = app.post('/', json_input_body,
                                  content_type='application/json; charset=utf8')
        self.assert_response_ok(resp)

    def test_json_with_suffix(self):
        self.assert_json_ok(suffix='Request')

    def test_json_no_suffix(self):
        self.assert_json_ok(suffix='')

    def test_json_with_suffix_with_message_name(self):
        self.assert_json_ok(suffix='Request',
                            _in_message_name='TestInMessageName')

    def test_json_no_suffix_with_message_name(self):
        self.assert_json_ok(suffix='', _in_message_name='TestInMessageName')

    def test_json_with_suffix_with_operation_name(self):
        self.assert_soap_ok(suffix='Request',
                            _operation_name='TestOperationName')

    def test_json_no_suffix_with_operation_name(self):
        self.assert_soap_ok(suffix='', _operation_name='TestOperationName')

    ### HttpRpc tests ###
    def assert_httprpc_ok(self, suffix, _operation_name=None,
                          _in_message_name=None):
        """Helper to test HttpRpc requests"""

        # setup
        app = self.get_app(HttpRpc(validator='soft'),
                           suffix, _operation_name, _in_message_name)

        function_name, operation_name, request_name = \
            self.get_function_names(suffix, _operation_name, _in_message_name)

        url = "/{0}?string=test&times=2".format(request_name)

        # check httprpc operation succeeded
        resp = app.get(url)
        self.assert_response_ok(resp)

    def test_httprpc_with_suffix(self):
        self.assert_httprpc_ok(suffix='Request')

    def test_httprpc_no_suffix(self):
        self.assert_httprpc_ok(suffix='')

    def test_httprpc_with_suffix_with_message_name(self):
        self.assert_httprpc_ok(suffix='Request',
                               _in_message_name='TestInMessageName')

    def test_httprpc_no_suffix_with_message_name(self):
        self.assert_httprpc_ok(suffix='', _in_message_name='TestInMessageName')

    def test_httprpc_with_suffix_with_operation_name(self):
        self.assert_soap_ok(suffix='Request',
                            _operation_name='TestOperationName')

    def test_httprpc_no_suffix_with_operation_name(self):
        self.assert_soap_ok(suffix='', _operation_name='TestOperationName')




#
# spyne - Copyright (C) Spyne contributors.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
#

from spyne.context import FakeContext

from spyne.protocol.dictdoc import HierDictDocument
from spyne.protocol.dictdoc import SimpleDictDocument

try:
    from spyne.protocol.json import JsonDocument
except ImportError as _import_error:
    _local_import_error = _import_error
    def JsonDocument(*args, **kwargs):
        raise _local_import_error


try:
    from spyne.protocol.yaml import YamlDocument
except ImportError as _import_error:
    _local_import_error = _import_error
    def YamlDocument(*args, **kwargs):
        raise _local_import_error


try:
    from spyne.protocol.msgpack import MessagePackDocument
except ImportError as _import_error:
    _local_import_error = _import_error
    def MessagePackDocument(*args, **kwargs):
        raise _local_import_error


from spyne.model.primitive import Double
from spyne.model.primitive import Boolean
from spyne.model.primitive import Decimal
from spyne.model.primitive import Integer


class _UtilProtocol(HierDictDocument):
    def __init__(self, app=None, validator=None, mime_type=None,
                                        ignore_uncap=False,
                                        # DictDocument specific
                                        ignore_wrappers=True,
                                        complex_as=dict,
                                        ordered=False):

        super(_UtilProtocol, self).__init__(app, validator, mime_type, ignore_uncap,
                                           ignore_wrappers, complex_as, ordered)

        self._from_unicode_handlers[Double] = lambda cls, val: val
        self._from_unicode_handlers[Boolean] = lambda cls, val: val
        self._from_unicode_handlers[Decimal] = lambda cls, val: val
        self._from_unicode_handlers[Integer] = lambda cls, val: val

        self._to_unicode_handlers[Double] = lambda cls, val: val
        self._to_unicode_handlers[Boolean] = lambda cls, val: val
        self._to_unicode_handlers[Decimal] = lambda cls, val: val
        self._to_unicode_handlers[Integer] = lambda cls, val: val


def get_doc_as_object(d, cls, ignore_wrappers=True, complex_as=list,
                                    protocol=_UtilProtocol, protocol_inst=None):
    if protocol_inst is None:
        protocol_inst = protocol(ignore_wrappers=ignore_wrappers,
                                                          complex_as=complex_as)

    return protocol_inst._doc_to_object(None, cls, d)


get_dict_as_object = get_doc_as_object
"""DEPRECATED: Use ``get_doc_as_object`` instead"""


def get_object_as_doc(o, cls=None, ignore_wrappers=True, complex_as=dict,
                                    protocol=_UtilProtocol, protocol_inst=None):
    if cls is None:
        cls = o.__class__

    if protocol_inst is None:
        protocol_inst = protocol(ignore_wrappers=ignore_wrappers,
                                                          complex_as=complex_as)

    retval = protocol_inst._object_to_doc(cls, o)

    if not ignore_wrappers:
        return {cls.get_type_name(): retval}

    return retval


get_object_as_dict = get_object_as_doc
"""DEPRECATED: Use ``get_object_as_doc`` instead."""

def get_object_as_simple_dict(o, cls=None, hier_delim='.', prefix=None):
    if cls is None:
        cls = o.__class__

    return SimpleDictDocument(hier_delim=hier_delim) \
                                   .object_to_simple_dict(cls, o, prefix=prefix)


def get_object_as_json(o, cls=None, ignore_wrappers=True, complex_as=list,
                     encoding='utf8', polymorphic=False, indent=None, **kwargs):
    if cls is None:
        cls = o.__class__

    prot = JsonDocument(ignore_wrappers=ignore_wrappers, complex_as=complex_as,
                               polymorphic=polymorphic, indent=indent, **kwargs)
    ctx = FakeContext(out_document=[prot._object_to_doc(cls, o)])
    prot.create_out_string(ctx, encoding)
    return b''.join(ctx.out_string)


def get_object_as_json_doc(o, cls=None, ignore_wrappers=True, complex_as=list,
                                      polymorphic=False, indent=None, **kwargs):
    if cls is None:
        cls = o.__class__

    prot = JsonDocument(ignore_wrappers=ignore_wrappers, complex_as=complex_as,
                               polymorphic=polymorphic, indent=indent, **kwargs)

    return prot._object_to_doc(cls, o)


def get_object_as_yaml(o, cls=None, ignore_wrappers=False, complex_as=dict,
                                            encoding='utf8', polymorphic=False):
    if cls is None:
        cls = o.__class__

    prot = YamlDocument(ignore_wrappers=ignore_wrappers, complex_as=complex_as,
                                                        polymorphic=polymorphic)
    ctx = FakeContext(out_document=[prot._object_to_doc(cls,o)])
    prot.create_out_string(ctx, encoding)
    return b''.join(ctx.out_string)


def get_object_as_yaml_doc(o, cls=None, ignore_wrappers=False, complex_as=dict,
                                                             polymorphic=False):
    if cls is None:
        cls = o.__class__

    prot = YamlDocument(ignore_wrappers=ignore_wrappers, complex_as=complex_as,
                                                        polymorphic=polymorphic)
    return prot._object_to_doc(cls, o)


def get_object_as_msgpack(o, cls=None, ignore_wrappers=False, complex_as=dict,
                                                             polymorphic=False):
    if cls is None:
        cls = o.__class__

    prot = MessagePackDocument(ignore_wrappers=ignore_wrappers,
                                 complex_as=complex_as, polymorphic=polymorphic)
    ctx = FakeContext(out_document=[prot._object_to_doc(cls,o)])
    prot.create_out_string(ctx)
    return b''.join(ctx.out_string)


def get_object_as_msgpack_doc(o, cls=None, ignore_wrappers=False,
                                            complex_as=dict, polymorphic=False):
    if cls is None:
        cls = o.__class__

    prot = MessagePackDocument(ignore_wrappers=ignore_wrappers,
                                 complex_as=complex_as, polymorphic=polymorphic)

    return prot._object_to_doc(cls, o)


def json_loads(s, cls, protocol=JsonDocument, **kwargs):
    if s is None:
        return None
    if s == '':
        return None
    prot = protocol(**kwargs)
    ctx = FakeContext(in_string=[s])
    prot.create_in_document(ctx)
    return prot._doc_to_object(None, cls, ctx.in_document,
                                                       validator=prot.validator)


get_json_as_object = json_loads


def yaml_loads(s, cls, protocol=YamlDocument, ignore_wrappers=False, **kwargs):
    if s is None:
        return None
    if s == '' or s == b'':
        return None
    prot = protocol(ignore_wrappers=ignore_wrappers, **kwargs)
    ctx = FakeContext(in_string=[s])
    prot.create_in_document(ctx)
    retval = prot._doc_to_object(None, cls, ctx.in_document,
                                                       validator=prot.validator)
    return retval


get_yaml_as_object = yaml_loads



#!/usr/bin/env python

import sys

import spyne.const
spyne.const.MIN_GC_INTERVAL = float('inf')

from lxml import html

from random import randint, shuffle, choice
from contextlib import closing
from email.utils import formatdate

from neurons import TableModel, Application
from neurons.daemon import ServiceDefinition, HttpServer, StaticFileServer
from neurons.daemon.main import Bootstrapper

from spyne import Integer32, Unicode, rpc, ServiceBase, Integer, Array, Any
from spyne.protocol.html import HtmlCloth
from spyne.protocol.http import HttpRpc
from spyne.protocol.json import JsonDocument
from spyne.server.wsgi import WsgiApplication

if sys.version_info[0] == 3:
    xrange = range

_is_pypy = hasattr(sys, 'pypy_version_info')

DBDRIVER = 'postgresql+psycopg2cffi' if _is_pypy else 'postgresql+psycopg2'
DBHOST = 'tfb-database'


# models
class DbSessionManager(object):
    def __init__(self, config):
        self.session = config.get_main_store().Session()

    def close(self, with_err):
        self.session.close()


class DbConnectionManager(object):
    def __init__(self, config):
        self.conn = config.get_main_store().engine.connect()

    def close(self, with_err):
        self.conn.close()


class World(TableModel):
    __tablename__ = "world"
    _type_info = [
        ('id', Integer32(primary_key=True)),
        ('randomNumber', Integer32(sqla_column_args=dict(name='randomnumber'))),
    ]


T_INDEX = html.fromstring(open('cloths/index.html', 'rb').read())


class Fortune(TableModel):
    __tablename__ = "fortune"

    id = Integer32(primary_key=True)
    message = Unicode


outprot_plain = HttpRpc(mime_type='text/plain')


class TfbSimpleService(ServiceBase):
    @rpc(_returns=Any)
    def json(ctx):
        ctx.transport.add_header('Date', formatdate(usegmt=True))
        return dict(message=u'Hello, World!')

    @rpc(_returns=Any)
    def plaintext(ctx):
        """Test 6: Plaintext"""
        ctx.out_protocol = outprot_plain
        return b'Hello, World!'


def _force_int(v):
    try:
        return min(500, max(int(v), 1))
    except:
        return 1


NumQueriesType = Any(sanitizer=_force_int)


class TfbOrmService(ServiceBase):
    @rpc(_returns=World)
    def db(ctx):
        retval = ctx.udc.session.query(World).get(randint(1, 10000))
        return retval

    @rpc(NumQueriesType, _returns=Array(World))
    def dbs(ctx, queries):
        if queries is None:
            queries = 1

        q = ctx.udc.session.query(World)
        return [q.get(randint(1, 10000)) for _ in xrange(queries)]

    @rpc(_returns=Array(Fortune, html_cloth=T_INDEX), _body_style='out_bare')
    def fortunes(ctx):
        # This is normally specified at the application level as it's a good
        # practice to group rpc endpoints with similar return types under the
        # same url fragment. eg. https://example.com/api/json
        ctx.out_protocol = HtmlCloth()
        ctx.outprot_ctx = ctx.out_protocol.get_context(ctx, ctx.transport)

        fortunes = ctx.udc.session.query(Fortune).all()
        fortunes.append(
            Fortune(id=0, message=u"Additional fortune added at request time.")
        )

        fortunes.sort(key=lambda x: x.message)

        return fortunes

    @rpc(NumQueriesType, _returns=Array(World))
    def updates(ctx, queries):
        """Test 5: Database Updates"""

        if queries is None:
            queries = 1

        retval = []
        q = ctx.udc.session.query(World)
        for id in (randint(1, 10000) for _ in xrange(queries)):
            world = q.get(id)
            world.randomNumber = randint(1, 10000)
            retval.append(world)

        ctx.udc.session.commit()

        return retval


class TfbRawService(ServiceBase):
    @rpc(_returns=World)
    def dbraw(ctx):
        conn = ctx.udc.conn

        wid = randint(1, 10000)
        return conn.execute(
                     "SELECT id, randomNumber FROM world WHERE id = %s", wid) \
                                                                     .fetchone()

    # returning both Any+dict or ObjectMarker+ListOfLists works
    @rpc(NumQueriesType, _returns=Any)
    def dbsraw(ctx, queries):
        if queries is None:
            queries = 1

        retval = []
        conn = ctx.udc.conn
        for i in xrange(queries):
            wid = randint(1, 10000)
            result = conn.execute(
                     "SELECT id, randomNumber FROM world WHERE id = %s", wid) \
                .fetchone()
            retval.append(dict(id=result[0], randomNumber=result[1]))

        return retval

    @rpc(_returns=Array(Fortune, html_cloth=T_INDEX), _body_style='out_bare')
    def fortunesraw(ctx):
        # This is normally specified at the application level as it's a good
        # practice to group rpc endpoints with similar return types under the
        # same url fragment. eg. https://example.com/api/json
        ctx.out_protocol = HtmlCloth()
        ctx.outprot_ctx = ctx.out_protocol.get_context(ctx, ctx.transport)

        res = ctx.udc.conn.execute("SELECT id, message FROM fortune")
        fortunes = res.fetchall()

        fortunes.append(Fortune(
            id=0,
            message=u"Additional fortune added at request time."
        ))

        fortunes.sort(key=lambda x: x.message)

        return fortunes

    @rpc(NumQueriesType, _returns=Any)
    def updatesraw(ctx, queries):
        """Test 5: Database Updates"""
        if queries is None:
            queries = 1

        conn = ctx.udc.conn

        ids = [randint(1, 10000) for _ in xrange(queries)]

        retval = []
        for i in ids:
            wid, rn = conn.execute(
                         "SELECT id, randomNumber FROM world WHERE id=%s", i) \
                .fetchone()

            rn = randint(1, 10000)
            retval.append(dict(id=wid, randomNumber=rn))

            conn.execute("UPDATE World SET randomNumber=%s WHERE id=%s",
                                                                        rn, wid)

        return retval


def _on_method_call_db_sess(ctx):
    ctx.transport.add_header('Date', formatdate(usegmt=True))
    ctx.udc = DbSessionManager(ctx.app.config)


def _on_method_call_db_conn(ctx):
    ctx.transport.add_header('Date', formatdate(usegmt=True))
    ctx.udc = DbConnectionManager(ctx.app.config)


TfbRawService.event_manager.add_listener("method_call", _on_method_call_db_conn)
TfbOrmService.event_manager.add_listener("method_call", _on_method_call_db_sess)


def init_app(config):
    subconfig = config.services['root']

    app = Application(
        [TfbOrmService, TfbRawService, TfbSimpleService],
        tns='http://techempower.com/benchmarks/Python/Spyne',
        in_protocol=HttpRpc(),
        out_protocol=JsonDocument(),
        config=config,
    )
    if subconfig.subapps is None:
        subconfig.subapps = {}

    subconfig.subapps.update({'': app})

    return subconfig.gen_site()


def init(config):
    return {
        'root': ServiceDefinition(
            init=init_app,
            default=HttpServer(
                type='tcp4',
                host='127.0.0.1',
                port=8080,
            ),
        ),
    }


def parse_config(argv):
    from neurons.daemon.main import boot

    retcode, config = boot('tfb', argv, init, bootstrapper=TfbBootstrap)

    if retcode is not None:
        sys.exit(retcode)

    return config


def gen_wsgi_app():
    config = parse_config([])
    app = config.services['root'].subapps[''].app
    return WsgiApplication(app)


words = 'some random words for you and me somebody else if then the'.split()


class TfbBootstrap(Bootstrapper):
    # noinspection PyUnresolvedReferences
    def after_tables(self, config):
        print("Generating data...")
        with closing(config.get_main_store().Session()) as session:
            ints = list(range(10000))
            shuffle(ints)
            for _ in range(10000):
                session.add(World(randomNumber=ints.pop()))

            for _ in range(100):
                session.add(Fortune(
                    message=' '.join([choice(words)
                                                for _ in range(randint(3, 10))])
                ))

            session.commit()


if __name__ == '__main__':
    parse_config(sys.argv)
else:
    application = gen_wsgi_app()




#
# spyne - Copyright (C) Spyne contributors.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
#


"""The ``spyne.protocol.xml`` module contains an xml-based protocol that
serializes python objects to xml using Xml Schema conventions.

Logs valid documents to ``'spyne.protocol.xml'`` and invalid documents to
``spyne.protocol.xml.invalid``. Use the usual ``logging.getLogger()`` and
friends to configure how these get logged.

Warning! You can get a lot of crap in the 'invalid' logger. You're not advised
to turn it on for a production system.
"""


import logging
logger = logging.getLogger('spyne.protocol.xml')
logger_invalid = logging.getLogger('spyne.protocol.xml.invalid')

from inspect import isgenerator
from collections import defaultdict

from lxml import etree
from lxml import html
from lxml.builder import E
from lxml.etree import XMLSyntaxError
from lxml.etree import XMLParser

from spyne import BODY_STYLE_WRAPPED

from spyne.util import Break, coroutine
from spyne.util.six import text_type, string_types
from spyne.util.cdict import cdict
from spyne.util.etreeconv import etree_to_dict, dict_to_etree,\
    root_dict_to_etree
from spyne.const.xml import XSI, NS_SOAP11_ENC

from spyne.error import Fault
from spyne.error import ValidationError
from spyne.const.ansi_color import LIGHT_GREEN
from spyne.const.ansi_color import LIGHT_RED
from spyne.const.ansi_color import END_COLOR
from spyne.const.xml import NS_SOAP11_ENV
from spyne.const.xml import PREFMAP, DEFAULT_NS

from spyne.model import Any, ModelBase, Array, Iterable, ComplexModelBase, \
    AnyHtml, AnyXml, AnyDict, Unicode, PushBase, File, ByteArray, XmlData, \
    XmlAttribute
from spyne.model.binary import BINARY_ENCODING_BASE64
from spyne.model.enum import EnumBase

from spyne.protocol import ProtocolBase

from spyne.util import six

if six.PY2:
    STR_TYPES = (str, unicode)
else:
    STR_TYPES = (str, bytes)


NIL_ATTR = {XSI('nil'): 'true'}
XSI_TYPE = XSI('type')


def _append(parent, child_elt):
    if hasattr(parent, 'append'):
        parent.append(child_elt)
    else:
        parent.write(child_elt)


def _gen_tagname(ns, name):
    if ns is not None:
        name = "{%s}%s" % (ns, name)
    return name


class SchemaValidationError(Fault):
    """Raised when the input stream could not be validated by the Xml Schema."""

    CODE = 'Client.SchemaValidationError'

    def __init__(self, faultstring):
        super(SchemaValidationError, self).__init__(self.CODE, faultstring)


class SubXmlBase(ProtocolBase):
    def subserialize(self, ctx, cls, inst, parent, ns=None, name=None):
        return self.to_parent(ctx, cls, inst, parent, name)

    def to_parent(self, ctx, cls, inst, parent, ns, *args, **kwargs):
        """Serializes inst to an Element instance and appends it to the 'parent'.

        :param self:  The protocol that will be used to serialize the given
            value.
        :param cls:   The type of the value that's going to determine how to
            pack the given value.
        :param inst: The value to be set for the 'text' element of the newly
            created SubElement
        :param parent: The parent Element to which the new child will be
            appended.
        :param ns:   The target namespace of the new SubElement, used with
            'name' to set the tag.
        :param name:  The tag name of the new SubElement, 'retval' by default.
        """
        raise NotImplementedError()


class XmlDocument(SubXmlBase):
    """The Xml input and output protocol, using the information from the Xml
    Schema generated by Spyne types.

    See the following material for more (much much more!) information.

    * http://www.w3.org/TR/xmlschema-0/
    * http://www.w3.org/TR/xmlschema-1/
    * http://www.w3.org/TR/xmlschema-2/

    Receiving Xml from untrusted sources is a dodgy security dance as the Xml
    attack surface is /huge/.

    Spyne's ```lxml.etree.XMLParser``` instance has ```resolve_pis```,
    ```load_dtd```, ```resolve_entities```, ```dtd_validation```,
    ```huge_tree``` Defaults to ``False``

    Having ```resolve_entities``` disabled will prevent the 'lxml' validation
    for documents with custom xml entities defined in the DTD. See the example
    in examples/xml/validation_error to play with the settings that work best
    for you. Please note that enabling ```resolve_entities``` is a security
    hazard that can lead to disclosure of sensitive information.

    See https://pypi.python.org/pypi/defusedxml for a pragmatic overview of
    Xml security in Python world.

    :param app: The owner application instance.

    :param validator: One of (None, 'soft', 'lxml', 'schema',
        ProtocolBase.SOFT_VALIDATION, XmlDocument.SCHEMA_VALIDATION).
        Both ``'lxml'`` and ``'schema'`` values are equivalent to
        ``XmlDocument.SCHEMA_VALIDATION``.

        Defaults to ``None``.

    :param replace_null_with_default: If ``False``, does not replace incoming
        explicit null values with denoted default values. This is against Xml
        Schema standard but consistent with other Spyne protocol
        implementations. Set this to False if you want cross-protocol
        compatibility.

        Defaults to ``True``.

        Relevant quote from xml schema primer
        (http://www.w3.org/TR/xmlschema-0/):

        ..
            When a value is present and is null The schema processor treats
            defaulted elements slightly differently. When an element is declared
            with a default value, the value of the element is whatever value
            appears as the element's content in the instance document; if the
            element appears without any content, the schema processor provides
            the element with a value equal to that of the default attribute.
            However, if the element does not appear in the instance document,
            the schema processor does not provide the element at all. In
            summary, the differences between element and attribute defaults can
            be stated as: Default attribute values apply when attributes are
            missing, and default element values apply when elements are empty.

    :param xml_declaration: Whether to add xml_declaration to the responses

        Defaults to ``True``.

    :param cleanup_namespaces: Whether to add clean up namespace declarations
        in the response document.

        Defaults to ``True``.

    :param encoding: The suggested string encoding for the returned xml
        documents. The transport can override this.

        Defaults to ``None``.

    :param pretty_print: When ``True``, returns the document in a pretty-printed
        format.

        Defaults to ``False``.

    :param parse_xsi_type: Set to ``False`` to disable parsing of ``xsi:type``
        attribute, effectively disabling polymorphism.

        Defaults to ``True``.

    The following are passed straight to the ``XMLParser()`` instance from
    lxml. Docs are also plagiarized from the lxml documentation. Please note
    that some of the defaults are different to make parsing safer by default.

    :param attribute_defaults: read the DTD (if referenced by the document) and
        add the default attributes from it.

        Defaults to ``False``

    :param dtd_validation: validate while parsing (if a DTD was referenced).

        Defaults to ``False``

    :param load_dtd: load and parse the DTD while parsing (no validation is
        performed).

        Defaults to ``False``.

    :param no_network: prevent network access when looking up external
        documents.

        Defaults to ``True``.

    :param ns_clean: try to clean up redundant namespace declarations.
        Please note that this is for incoming documents.
        See ``cleanup_namespaces`` parameter for output documents.

        Defaults to ``False``.

    :param recover: try hard to parse through broken Xml.

        Defaults to ``False``.

    :param remove_blank_text: discard blank text nodes between tags, also known
        as ignorable whitespace. This is best used together with a DTD or schema
        (which tells data and noise apart), otherwise a heuristic will be
        applied.

        Defaults to ``False``.

    :param remove_pis: When ``True`` xml parser discards processing
        instructions.

        Defaults to ``True``.

    :param strip_cdata: replace CDATA sections by normal text content.

        Defaults to ``True``

    :param resolve_entities: replace entities by their text value.

        Defaults to ``False``.

    :param huge_tree: disable security restrictions and support very deep trees
        and very long text content. (only affects libxml2 2.7+)

        Defaults to ``False``.

    :param compact: use compact storage for short text content.

        Defaults to ``True``.

    """

    SCHEMA_VALIDATION = type("Schema", (object,), {})

    mime_type = 'text/xml'
    default_binary_encoding = BINARY_ENCODING_BASE64

    type = set(ProtocolBase.type)
    type.add('xml')

    soap_env = PREFMAP[NS_SOAP11_ENV]
    ns_soap_env = NS_SOAP11_ENV
    ns_soap_enc = NS_SOAP11_ENC

    def __init__(self, app=None, validator=None,
                replace_null_with_default=True,
                xml_declaration=True,
                cleanup_namespaces=True, encoding=None, pretty_print=False,
                attribute_defaults=False,
                dtd_validation=False,
                load_dtd=False,
                no_network=True,
                ns_clean=False,
                recover=False,
                remove_blank_text=False,
                remove_pis=True,
                strip_cdata=True,
                resolve_entities=False,
                huge_tree=False,
                compact=True,
                binary_encoding=None,
                parse_xsi_type=True,
                polymorphic=False,
            ):

        super(XmlDocument, self).__init__(app, validator,
                                                binary_encoding=binary_encoding)

        self.validation_schema = None
        self.xml_declaration = xml_declaration
        self.cleanup_namespaces = cleanup_namespaces
        self.replace_null_with_default = replace_null_with_default

        if encoding is None:
            self.encoding = 'UTF-8'
        else:
            self.encoding = encoding

        self.polymorphic = polymorphic
        self.pretty_print = pretty_print
        self.parse_xsi_type = parse_xsi_type

        self.serialization_handlers = cdict({
            Any: self.any_to_parent,
            Fault: self.fault_to_parent,
            EnumBase: self.enum_to_parent,
            AnyXml: self.any_xml_to_parent,
            XmlData: self.xmldata_to_parent,
            AnyDict: self.any_dict_to_parent,
            AnyHtml: self.any_html_to_parent,
            ModelBase: self.modelbase_to_parent,
            ByteArray: self.byte_array_to_parent,
            ComplexModelBase: self.complex_to_parent,
            XmlAttribute: self.xmlattribute_to_parent,
            SchemaValidationError: self.schema_validation_error_to_parent,
        })

        self.deserialization_handlers = cdict({
            AnyHtml: self.html_from_element,
            AnyXml: self.xml_from_element,
            Any: self.xml_from_element,
            Array: self.array_from_element,
            Fault: self.fault_from_element,
            AnyDict: self.dict_from_element,
            EnumBase: self.enum_from_element,
            ModelBase: self.base_from_element,
            Unicode: self.unicode_from_element,
            Iterable: self.iterable_from_element,
            ByteArray: self.byte_array_from_element,
            ComplexModelBase: self.complex_from_element,
        })

        self.parser_kwargs = dict(
            attribute_defaults=attribute_defaults,
            dtd_validation=dtd_validation,
            load_dtd=load_dtd,
            no_network=no_network,
            ns_clean=ns_clean,
            recover=recover,
            remove_blank_text=remove_blank_text,
            remove_comments=True,
            remove_pis=remove_pis,
            strip_cdata=strip_cdata,
            resolve_entities=resolve_entities,
            huge_tree=huge_tree,
            compact=compact,
            encoding=encoding,
        )

    def set_validator(self, validator):
        if validator in ('lxml', 'schema') or \
                                    validator is self.SCHEMA_VALIDATION:
            self.validate_document = self.__validate_lxml
            self.validator = self.SCHEMA_VALIDATION

        elif validator == 'soft' or validator is self.SOFT_VALIDATION:
            self.validator = self.SOFT_VALIDATION

        elif validator is None:
            pass

        else:
            raise ValueError(validator)

        self.validation_schema = None

    def validate_body(self, ctx, message):
        """Sets ctx.method_request_string and calls :func:`generate_contexts`
        for validation."""

        assert message in (self.REQUEST, self.RESPONSE), message

        line_header = LIGHT_RED + "Error:" + END_COLOR
        try:
            self.validate_document(ctx.in_body_doc)
            if message is self.REQUEST:
                line_header = LIGHT_GREEN + "Method request string:" + END_COLOR
            else:
                line_header = LIGHT_RED + "Response:" + END_COLOR
        finally:
            if logger.level == logging.DEBUG:
                logger.debug("%s %s" % (line_header, ctx.method_request_string))
                logger.debug(etree.tostring(ctx.in_document, pretty_print=True))

    def set_app(self, value):
        ProtocolBase.set_app(self, value)

        self.validation_schema = None

        if self.validator is self.SCHEMA_VALIDATION and value is not None:
            xml_schema = self.app.interface.docs.xml_schema
            xml_schema.build_validation_schema()
            self.validation_schema = xml_schema.validation_schema

    def __validate_lxml(self, payload):
        ret = self.validation_schema.validate(payload)

        logger.debug("Validated ? %r" % ret)
        if ret == False:
            error_text = text_type(self.validation_schema.error_log.last_error)
            raise SchemaValidationError(error_text.encode('ascii',
                                                           'xmlcharrefreplace'))

    def create_in_document(self, ctx, charset=None):
        """Uses the iterable of string fragments in ``ctx.in_string`` to set
        ``ctx.in_document``."""

        string = b''.join(ctx.in_string)
        try:
            try:
                ctx.in_document = etree.fromstring(string,
                                        parser=XMLParser(**self.parser_kwargs))

            except ValueError:
                logger.debug('ValueError: Deserializing from unicode strings '
                             'with encoding declaration is not supported by '
                             'lxml.')
                ctx.in_document = etree.fromstring(string.decode(charset),
                                                                    self.parser)
        except XMLSyntaxError as e:
            logger_invalid.error("%r in string %r", e, string)
            raise Fault('Client.XMLSyntaxError', str(e))

    def decompose_incoming_envelope(self, ctx, message):
        assert message in (self.REQUEST, self.RESPONSE)

        ctx.in_header_doc = None # If you need header support, you should use Soap
        ctx.in_body_doc = ctx.in_document
        ctx.method_request_string = ctx.in_body_doc.tag
        self.validate_body(ctx, message)

    def from_element(self, ctx, cls, element):
        cls_attrs = self.get_cls_attrs(cls)

        if bool(element.get(XSI('nil'))):
            if self.validator is self.SOFT_VALIDATION and not \
                                                             cls_attrs.nillable:
                raise ValidationError(None)

            if self.replace_null_with_default:
                return cls_attrs.default

            return None

        # if present, use the xsi:type="ns0:ObjectName"
        # attribute to instantiate subclass objects
        if self.parse_xsi_type:
            xsi_type = element.get(XSI_TYPE, None)
            if xsi_type is not None:
                if ":" in xsi_type:
                    prefix, objtype = xsi_type.split(':', 1)
                else:
                    prefix, objtype = None, xsi_type

                ns = element.nsmap.get(prefix)
                if ns is not None:
                    classkey = "{%s}%s" % (ns, objtype)

                else:
                    logger.error("xsi:type namespace prefix "
                                            "'%s' in '%s' not found in %r",
                                                    ns, xsi_type, element.nsmap)

                    raise ValidationError(xsi_type)

                newclass = ctx.app.interface.classes.get(classkey, None)
                if newclass is None:
                    logger.error("xsi:type '%s' interpreted as class key '%s' "
                                        "is not recognized", xsi_type, classkey)
                    raise ValidationError(xsi_type)

                cls = newclass
                logger.debug("xsi:type '%s' overrides %r to %r", xsi_type,
                                                                  cls, newclass)

        handler = self.deserialization_handlers[cls]
        return handler(ctx, cls, element)

    def to_parent(self, ctx, cls, inst, parent, ns, *args, **kwargs):
        cls, add_type = self.get_polymorphic_target(cls, inst)
        cls_attrs = self.get_cls_attrs(cls)

        subprot = cls_attrs.prot
        if subprot is not None and isinstance(subprot, SubXmlBase):
            return subprot.subserialize(ctx, cls, inst, parent, ns,
                                                                *args, **kwargs)

        handler = self.serialization_handlers[cls]

        if inst is None:
            inst = cls_attrs.default

        if inst is None:
            return self.null_to_parent(ctx, cls, inst, parent, ns,
                                                                *args, **kwargs)

        if cls_attrs.exc:
            return

        kwargs['add_type'] = add_type
        return handler(ctx, cls, inst, parent, ns, *args, **kwargs)

    def deserialize(self, ctx, message):
        """Takes a MethodContext instance and a string containing ONE root xml
        tag.

        Returns the corresponding native python object.

        Not meant to be overridden.
        """

        assert message in (self.REQUEST, self.RESPONSE)

        self.event_manager.fire_event('before_deserialize', ctx)

        if ctx.descriptor is None:
            if ctx.in_error is None:
                raise Fault("Client", "Method %r not found." %
                                                      ctx.method_request_string)
            else:
                raise ctx.in_error

        if message is self.REQUEST:
            body_class = ctx.descriptor.in_message
        elif message is self.RESPONSE:
            body_class = ctx.descriptor.out_message

        # decode method arguments
        if ctx.in_body_doc is None:
            ctx.in_object = [None] * len(body_class._type_info)
        else:
            ctx.in_object = self.from_element(ctx, body_class, ctx.in_body_doc)

        if logger.level == logging.DEBUG and message is self.REQUEST:
            line_header = '%sRequest%s' % (LIGHT_GREEN, END_COLOR)

            outdoc_str = None
            if ctx.out_document is not None:
                outdoc_str = etree.tostring(ctx.out_document,
                        xml_declaration=self.xml_declaration, pretty_print=True)

            logger.debug("%s %s" % (line_header, outdoc_str))

        self.event_manager.fire_event('after_deserialize', ctx)

    def serialize(self, ctx, message):
        """Uses ``ctx.out_object``, ``ctx.out_header`` or ``ctx.out_error`` to
        set ``ctx.out_body_doc``, ``ctx.out_header_doc`` and
        ``ctx.out_document`` as an ``lxml.etree._Element instance``.

        Not meant to be overridden.
        """

        assert message in (self.REQUEST, self.RESPONSE)

        self.event_manager.fire_event('before_serialize', ctx)

        if ctx.out_error is not None:
            tmp_elt = etree.Element('punk')
            retval = self.to_parent(ctx, ctx.out_error.__class__, ctx.out_error,
                                    tmp_elt, self.app.interface.get_tns())
            ctx.out_document = tmp_elt[0]

        else:
            if message is self.REQUEST:
                result_message_class = ctx.descriptor.in_message
            elif message is self.RESPONSE:
                result_message_class = ctx.descriptor.out_message

            # assign raw result to its wrapper, result_message
            if ctx.descriptor.body_style == BODY_STYLE_WRAPPED:
                result_inst = result_message_class()

                for i, (k, v) in enumerate(
                                       result_message_class._type_info.items()):
                    attrs = self.get_cls_attrs(v)
                    result_inst._safe_set(k, ctx.out_object[i], v, attrs)

            else:
                result_inst = ctx.out_object

            if ctx.out_stream is None:
                tmp_elt = etree.Element('punk')
                retval = self.to_parent(ctx, result_message_class,
                          result_inst, tmp_elt, self.app.interface.get_tns())
                ctx.out_document = tmp_elt[0]

            else:
                retval = self.incgen(ctx, result_message_class,
                                  result_inst, self.app.interface.get_tns())

        if self.cleanup_namespaces and ctx.out_document is not None:
            etree.cleanup_namespaces(ctx.out_document)

        self.event_manager.fire_event('after_serialize', ctx)

        return retval

    def create_out_string(self, ctx, charset=None):
        """Sets an iterable of string fragments to ctx.out_string"""

        if charset is None:
            charset = self.encoding

        ctx.out_string = [etree.tostring(ctx.out_document,
                                          encoding=charset,
                                          pretty_print=self.pretty_print,
                                          xml_declaration=self.xml_declaration)]

        if logger.level == logging.DEBUG:
            logger.debug('%sResponse%s %s' % (LIGHT_RED, END_COLOR,
                            etree.tostring(ctx.out_document,
                                          pretty_print=True, encoding='UTF-8')))

    @coroutine
    def incgen(self, ctx, cls, inst, ns, name=None):
        if name is None:
            name = cls.get_type_name()
        with etree.xmlfile(ctx.out_stream) as xf:
            ret = self.to_parent(ctx, cls, inst, xf, ns, name)
            if isgenerator(ret):
                try:
                    while True:
                        y = (yield) # may throw Break
                        ret.send(y)

                except Break:
                    try:
                        ret.throw(Break())
                    except StopIteration:
                        pass

        if hasattr(ctx.out_stream, 'finish'):
            ctx.out_stream.finish()

    def _gen_tag(self, cls, ns, name, add_type=False, **_):
        if ns is not None:
            name = "{%s}%s" % (ns, name)

        retval = E(name)
        if add_type:
            retval.attrib[XSI_TYPE] = cls.get_type_name_ns(self.app.interface)

        return retval

    def byte_array_to_parent(self, ctx, cls, inst, parent, ns, name='retval',
                                                                      **kwargs):
        elt = self._gen_tag(cls, ns, name, **kwargs)
        elt.text = self.to_unicode(cls, inst, self.binary_encoding)
        _append(parent, elt)

    def modelbase_to_parent(self, ctx, cls, inst, parent, ns, name='retval',
                                                                      **kwargs):
        elt = self._gen_tag(cls, ns, name, **kwargs)
        elt.text = self.to_unicode(cls, inst)
        _append(parent, elt)

    def null_to_parent(self, ctx, cls, inst, parent, ns, name='retval',
                                                                      **kwargs):
        if issubclass(cls, XmlAttribute):
            return

        elif issubclass(cls, XmlData):
            parent.attrib.update(NIL_ATTR)

        else:
            elt = self._gen_tag(cls, ns, name, **kwargs)
            elt.attrib.update(NIL_ATTR)
            _append(parent, elt)

    def null_from_element(self, ctx, cls, element):
        return None

    def xmldata_to_parent(self, ctx, cls, inst, parent, ns, name,
                                                      add_type=False, **_):
        cls_attrs = self.get_cls_attrs(cls)

        ns = cls._ns
        if ns is None:
            ns = cls_attrs.sub_ns

        name = _gen_tagname(ns, name)

        if add_type:
            parent.attrib[XSI_TYPE] = cls.get_type_name_ns(self.app.interface)

        cls.marshall(self, name, inst, parent)

    def xmlattribute_to_parent(self, ctx, cls, inst, parent, ns, name, **_):
        ns = cls._ns
        cls_attrs = self.get_cls_attrs(cls)
        if ns is None:
            ns = cls_attrs.sub_ns

        name = _gen_tagname(ns, name)

        if inst is not None:
            if issubclass(cls.type, (ByteArray, File)):
                parent.set(name, self.to_unicode(cls.type, inst,
                                                 self.binary_encoding))
            else:
                parent.set(name, self.to_unicode(cls.type, inst))

    @coroutine
    def gen_members_parent(self, ctx, cls, inst, parent, tag_name, subelts,
                                                                      add_type):
        attrib = {}
        if add_type:
            tnn = cls.get_type_name_ns(self.app.interface)
            if tnn != None:
                attrib[XSI_TYPE] = tnn
            else:
                # this only happens on incomplete interface states for eg.
                # get_object_as_xml where the full init is not performed for
                # perf reasons
                attrib[XSI_TYPE] = cls.get_type_name()

        if isinstance(parent, etree._Element):
            elt = etree.SubElement(parent, tag_name, attrib=attrib)
            elt.extend(subelts)
            ret = self._get_members_etree(ctx, cls, inst, elt)

            if isgenerator(ret):
                try:
                    while True:
                        y = (yield) # may throw Break
                        ret.send(y)

                except Break:
                    try:
                        ret.throw(Break())
                    except StopIteration:
                        pass

        else:
            with parent.element(tag_name, attrib=attrib):
                for e in subelts:
                    parent.write(e)
                ret = self._get_members_etree(ctx, cls, inst, parent)
                if isgenerator(ret):
                    try:
                        while True:
                            y = (yield)
                            ret.send(y)

                    except Break:
                        try:
                            ret.throw(Break())
                        except StopIteration:
                            pass

    @coroutine
    def _get_members_etree(self, ctx, cls, inst, parent):
        try:
            parent_cls = getattr(cls, '__extends__', None)

            if not (parent_cls is None):
                ret = self._get_members_etree(ctx, parent_cls, inst, parent)
                if ret is not None:
                    try:
                        while True:
                            sv2 = (yield) # may throw Break
                            ret.send(sv2)

                    except Break:
                        try:
                            ret.throw(Break())
                        except StopIteration:
                            pass

            for k, v in cls._type_info.items():
                sub_cls_attrs = self.get_cls_attrs(v)
                if sub_cls_attrs.exc:
                    continue

                try:
                    subvalue = getattr(inst, k, None)
                except:  # e.g. SqlAlchemy could throw NoSuchColumnError
                    subvalue = None

                # This is a tight loop, so enable this only when necessary.
                # logger.debug("get %r(%r) from %r: %r" % (k, v, inst, subvalue))

                sub_ns = v.Attributes.sub_ns
                if sub_ns is None:
                    sub_ns = cls.get_namespace()

                sub_name = v.Attributes.sub_name
                if sub_name is None:
                    sub_name = k

                mo = v.Attributes.max_occurs
                if subvalue is not None and mo > 1:
                    if isinstance(subvalue, PushBase):
                        while True:
                            sv = (yield)
                            ret = self.to_parent(ctx, v, sv, parent, sub_ns,
                                                                       sub_name)
                            if ret is not None:
                                try:
                                    while True:
                                        sv2 = (yield)  # may throw Break
                                        ret.send(sv2)

                                except Break:
                                    try:
                                        ret.throw(Break())
                                    except StopIteration:
                                        pass

                    else:
                        for sv in subvalue:
                            ret = self.to_parent(ctx, v, sv, parent, sub_ns,
                                                                       sub_name)

                            if ret is not None:
                                try:
                                    while True:
                                        sv2 = (yield)  # may throw Break
                                        ret.send(sv2)

                                except Break:
                                    try:
                                        ret.throw(Break())
                                    except StopIteration:
                                        pass

                # Don't include empty values for
                # non-nillable optional attributes.
                elif subvalue is not None or v.Attributes.min_occurs > 0:
                    ret = self.to_parent(ctx, v, subvalue, parent, sub_ns,
                                                                       sub_name)
                    if ret is not None:
                        try:
                            while True:
                                sv2 = (yield)
                                ret.send(sv2)
                        except Break as b:
                            try:
                                ret.throw(b)
                            except StopIteration:
                                pass

        except Break:
            pass

    def complex_to_parent(self, ctx, cls, inst, parent, ns, name=None,
                                                           add_type=False, **_):
        cls_attrs = self.get_cls_attrs(cls)

        sub_name = cls_attrs.sub_name
        if sub_name is not None:
            name = sub_name
        if name is None:
            name = cls.get_type_name()

        sub_ns = cls_attrs.sub_ns
        if not sub_ns in (None, DEFAULT_NS):
            ns = sub_ns

        tag_name = _gen_tagname(ns, name)

        inst = cls.get_serialization_instance(inst)

        return self.gen_members_parent(ctx, cls, inst, parent, tag_name, [],
                                                                       add_type)

    def _fault_to_parent_impl(self, ctx, cls, inst, parent, ns, subelts, **_):
        tag_name = "{%s}Fault" % self.ns_soap_env

        # Accepting raw lxml objects as detail is DEPRECATED. It's also not
        # documented. It's kept for backwards-compatibility purposes.
        if isinstance(inst.detail, string_types + (etree._Element,)):
            _append(subelts, E('detail', inst.detail))

        elif isinstance(inst.detail, dict):
            if len(inst.detail) > 0:
                _append(subelts, root_dict_to_etree({'detail':inst.detail}))

        elif inst.detail is None:
            pass

        else:
            raise TypeError('Fault detail Must be dict, got', type(inst.detail))

        # add other nonstandard fault subelements with get_members_etree
        return self.gen_members_parent(ctx, cls, inst, parent, tag_name,
                                                        subelts, add_type=False)

    def fault_to_parent(self, ctx, cls, inst, parent, ns, *args, **kwargs):
        subelts = [
            E("faultcode", '%s:%s' % (self.soap_env, inst.faultcode)),
            E("faultstring", inst.faultstring),
            E("faultactor", inst.faultactor),
        ]

        return self._fault_to_parent_impl(ctx, cls, inst, parent, ns, subelts)

    def schema_validation_error_to_parent(self, ctx, cls, inst, parent, ns,**_):
        subelts = [
            E("faultcode", '%s:%s' % (self.soap_env, inst.faultcode)),
            # HACK: Does anyone know a better way of injecting raw xml entities?
            E("faultstring", html.fromstring(inst.faultstring).text),
            E("faultactor", inst.faultactor),
        ]
        if inst.detail != None:
            _append(subelts, E('detail', inst.detail))

        # add other nonstandard fault subelements with get_members_etree
        return self._fault_to_parent_impl(ctx, cls, inst, parent, ns, subelts)

    def enum_to_parent(self, ctx, cls, inst, parent, ns, name='retval', **kwargs):
        self.modelbase_to_parent(ctx, cls, str(inst), parent, ns, name, **kwargs)

    def any_xml_to_parent(self, ctx, cls, inst, parent, ns, name, **_):
        if isinstance(inst, STR_TYPES):
            inst = etree.fromstring(inst)

        _append(parent, E(_gen_tagname(ns, name), inst))

    def any_to_parent(self, ctx, cls, inst, parent, ns, name, **_):
        _append(parent, E(_gen_tagname(ns, name), inst))

    def any_html_to_parent(self, ctx, cls, inst, parent, ns, name, **_):
        if isinstance(inst, string_types) and len(inst) > 0:
            inst = html.fromstring(inst)

        _append(parent, E(_gen_tagname(ns, name), inst))

    def any_dict_to_parent(self, ctx, cls, inst, parent, ns, name, **_):
        elt = E(_gen_tagname(ns, name))
        dict_to_etree(inst, elt)

        _append(parent, elt)

    def complex_from_element(self, ctx, cls, elt):
        inst = cls.get_deserialization_instance(ctx)

        flat_type_info = cls.get_flat_type_info(cls)

        # this is for validating cls.Attributes.{min,max}_occurs
        frequencies = defaultdict(int)
        cls_attrs = self.get_cls_attrs(cls)

        if cls_attrs._xml_tag_body_as is not None:
            for xtba_key, xtba_type in cls_attrs._xml_tag_body_as:
                xtba_attrs = self.get_cls_attrs(xtba_type.type)
                if issubclass(xtba_type.type, (ByteArray, File)):
                    value = self.from_unicode(xtba_type.type, elt.text,
                                                        self.binary_encoding)
                else:
                    value = self.from_unicode(xtba_type.type, elt.text)

                inst._safe_set(xtba_key, value, xtba_type.type, xtba_attrs)

        # parse input to set incoming data to related attributes.
        for c in elt:
            if isinstance(c, etree._Comment):
                continue

            key = c.tag.split('}', 1)[-1]
            frequencies[key] += 1

            member = flat_type_info.get(key, None)
            if member is None:
                member, key = cls._type_info_alt.get(key, (None, key))
                if member is None:
                    member, key = cls._type_info_alt.get(c.tag, (None, key))
                    if member is None:
                        continue

            member_attrs = self.get_cls_attrs(member)
            mo = member_attrs.max_occurs
            if mo > 1:
                value = getattr(inst, key, None)
                if value is None:
                    value = []

                value.append(self.from_element(ctx, member, c))

            else:
                value = self.from_element(ctx, member, c)

            inst._safe_set(key, value, member, member_attrs)

            for key, value_str in c.attrib.items():
                submember = flat_type_info.get(key, None)

                if submember is None:
                    submember, key = cls._type_info_alt.get(key, (None, key))
                    if submember is None:
                        continue

                submember_attrs = self.get_cls_attrs(submember)
                mo = submember_attrs.max_occurs
                if mo > 1:
                    value = getattr(inst, key, None)
                    if value is None:
                        value = []

                    value.append(self.from_unicode(submember.type, value_str))

                else:
                    value = self.from_unicode(submember.type, value_str)

                inst._safe_set(key, value, submember.type, submember_attrs)

        for key, value_str in elt.attrib.items():
            member = flat_type_info.get(key, None)
            if member is None:
                member, key = cls._type_info_alt.get(key, (None, key))
                if member is None:
                    continue

            if not issubclass(member, XmlAttribute):
                continue

            if issubclass(member.type, (ByteArray, File)):
                value = self.from_unicode(member.type, value_str,
                                                           self.binary_encoding)
            else:
                value = self.from_unicode(member.type, value_str)

            member_attrs = self.get_cls_attrs(member.type)
            inst._safe_set(key, value, member.type, member_attrs)

        if self.validator is self.SOFT_VALIDATION:
            for key, c in flat_type_info.items():
                val = frequencies.get(key, 0)
                attr = self.get_cls_attrs(c)
                if val < attr.min_occurs or val > attr.max_occurs:
                    raise Fault('Client.ValidationError', '%r member does not '
                                         'respect frequency constraints.' % key)

        return inst

    def array_from_element(self, ctx, cls, element):
        retval = [ ]
        (serializer,) = cls._type_info.values()

        for child in element.getchildren():
            retval.append(self.from_element(ctx, serializer, child))

        return retval

    def iterable_from_element(self, ctx, cls, element):
        (serializer,) = cls._type_info.values()

        for child in element.getchildren():
            yield self.from_element(ctx, serializer, child)

    def enum_from_element(self, ctx, cls, element):
        if self.validator is self.SOFT_VALIDATION and not (
                                        cls.validate_string(cls, element.text)):
            raise ValidationError(element.text)
        return getattr(cls, element.text)

    def fault_from_element(self, ctx, cls, element):
        code = element.find('faultcode').text
        string = element.find('faultstring').text
        factor = element.find('faultactor')
        if factor is not None:
            factor = factor.text
        detail = element.find('detail')

        return cls(faultcode=code, faultstring=string, faultactor=factor,
                                                                  detail=detail)

    def xml_from_element(self, ctx, cls, element):
        children = element.getchildren()
        retval = None

        if children:
            retval = element.getchildren()[0]

        return retval

    def html_from_element(self, ctx, cls, element):
        children = element.getchildren()
        retval = None

        if len(children) == 1:
            retval = children[0]
        # this is actually a workaround to a case that should never exist --
        # anyXml types should only have one child tag.
        elif len(children) > 1:
            retval = E.html(*children)

        return retval

    def dict_from_element(self, ctx, cls, element):
        children = element.getchildren()
        if children:
            return etree_to_dict(element)

        return None

    def unicode_from_element(self, ctx, cls, element):
        if self.validator is self.SOFT_VALIDATION and not (
                                        cls.validate_string(cls, element.text)):
            raise ValidationError(element.text)

        s = element.text
        if s is None:
            s = ''

        retval = self.from_unicode(cls, s)

        if self.validator is self.SOFT_VALIDATION and not (
                                              cls.validate_native(cls, retval)):
            raise ValidationError(retval)

        return retval

    def base_from_element(self, ctx, cls, element):
        if self.validator is self.SOFT_VALIDATION and not (
                                        cls.validate_string(cls, element.text)):
            raise ValidationError(element.text)

        retval = self.from_unicode(cls, element.text)

        if self.validator is self.SOFT_VALIDATION and not (
                                            cls.validate_native(cls, retval)):
            raise ValidationError(retval)

        return retval

    def byte_array_from_element(self, ctx, cls, element):
        if self.validator is self.SOFT_VALIDATION and not (
                                        cls.validate_string(cls, element.text)):
            raise ValidationError(element.text)

        retval = self.from_unicode(cls, element.text, self.binary_encoding)

        if self.validator is self.SOFT_VALIDATION and not (
                                            cls.validate_native(cls, retval)):
            raise ValidationError(retval)

        return retval



#!/usr/bin/env python
#
# spyne - Copyright (C) spyne contributors.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
#


import logging
logging.basicConfig(level=logging.DEBUG)

import unittest

from lxml import etree

from spyne.application import Application

from spyne.test.interface.wsdl import AppTestWrapper
from spyne.test.interface.wsdl import build_app
from spyne.test.interface.wsdl.defult_services import TDefaultPortService
from spyne.test.interface.wsdl.defult_services import \
                                              TDefaultPortServiceMultipleMethods

from spyne.const import REQUEST_SUFFIX
from spyne.const import RESPONSE_SUFFIX
from spyne.const import ARRAY_SUFFIX

from spyne.decorator import srpc
from spyne.service import Service
from spyne.interface.wsdl import Wsdl11
from spyne.model.complex import Array
from spyne.model.primitive import String

ns = {
    'wsdl': 'http://schemas.xmlsoap.org/wsdl/',
    'xs': 'http://www.w3.org/2001/XMLSchema',
}


class TestDefaultWSDLBehavior(unittest.TestCase):
    def _default_service(self, app_wrapper, service_name):
        self.assertEqual(1, len(app_wrapper.get_service_list()))

        services = app_wrapper.get_service_list()
        service = services[0]

        # the default behavior requires that there be only a single service
        self.assertEqual(1, len(services))
        self.assertEqual(service_name, service.get('name'))

        # Test the default service has the correct number of ports
        # the default behavior requires that there be only a single port
        ports = app_wrapper.get_port_list(service)
        self.assertEqual(len(ports), 1)

    def _default_port_type(self, app_wrapper, portType_name, op_count):
        # Verify the portType Count
        portTypes = app_wrapper.get_port_types()

        # there should be only one portType
        self.assertEqual(1, len(portTypes))

        # Verify the portType name
        portType = portTypes[0]
        # Check the name of the port
        self.assertEqual(portType_name, portType.get('name'))

        # verify that the portType definition has the correct
        # number of operations
        ops = app_wrapper.get_port_operations(portType)
        self.assertEqual(op_count, len(ops))

    def _default_binding(self, wrapper, binding_name, opp_count):
        # the default behavior is only single binding
        bindings = wrapper.get_bindings()
        self.assertEqual(1, len(bindings))

        # check for the correct binding name
        binding = bindings[0]
        name = binding.get('name')
        self.assertEqual(binding_name, name)

        # Test that the default service contains the soap binding
        sb = wrapper.get_soap_bindings(binding)
        self.assertEqual(1, len(sb))

        # verify the correct number of operations
        ops = wrapper.get_binding_operations(binding)
        self.assertEqual(opp_count, len(ops))

    def _default_binding_methods(self, wrapper, op_count, op_names):
        binding = wrapper.get_bindings()[0]
        operations = wrapper.get_binding_operations(binding)

        # Check the number of operations bound to the port
        self.assertEqual(op_count, len(operations))

        # Check the operation names are correct
        for op in operations:
            self.assertTrue(op.get('name') in op_names)

    def test_default_port_type(self):
        # Test the default port is created
        # Test the default port has the correct name
        app = build_app(
            [TDefaultPortService()],
            'DefaultPortTest',
            'DefaultPortName'
        )

        wrapper = AppTestWrapper(app)
        self._default_port_type(wrapper, 'DefaultPortName', 1)

    def test_default_port_type_multiple(self):
        app = build_app(
            [TDefaultPortServiceMultipleMethods()],
            'DefaultServiceTns',
            'MultipleDefaultPortServiceApp'
        )

        wrapper = AppTestWrapper(app)

        self._default_port_type(wrapper, "MultipleDefaultPortServiceApp", 3)

    def test_default_binding(self):
        app = build_app(
            [TDefaultPortService()],
            'DefaultPortTest',
            'DefaultBindingName'
        )

        wrapper = AppTestWrapper(app)

        self._default_binding(wrapper, "DefaultBindingName", 1)

    def test_default_binding_multiple(self):
        app = build_app(
            [TDefaultPortServiceMultipleMethods()],
            'DefaultPortTest',
            'MultipleDefaultBindingNameApp'
        )

        wrapper = AppTestWrapper(app)

        self._default_binding(wrapper, 'MultipleDefaultBindingNameApp', 3)

    def test_default_binding_methods(self):
        app = build_app(
            [TDefaultPortService()],
            'DefaultPortTest',
            'DefaultPortMethods'
        )

        wrapper = AppTestWrapper(app)

        self._default_binding_methods(
            wrapper,
            1,
            ['echo_default_port_service']
        )

    def test_bare_simple(self):
        class SomeService(Service):
            @srpc(String, _returns=String, _body_style='bare')
            def whatever(ss):
                return ss

        app = Application([SomeService], tns='tns')
        app.transport = 'None'

        wsdl = Wsdl11(app.interface)
        wsdl.build_interface_document('url')
        wsdl = etree.fromstring(wsdl.get_interface_document())

        schema = wsdl.xpath(
            '/wsdl:definitions/wsdl:types/xs:schema[@targetNamespace="tns"]',
            namespaces=ns,
        )
        assert len(schema) == 1

        print(etree.tostring(wsdl, pretty_print=True))

        elts = schema[0].xpath(
            'xs:element[@name="whatever%s"]' % REQUEST_SUFFIX, namespaces=ns)
        assert len(elts) > 0
        assert elts[0].attrib['type'] == 'xs:string'

        elts = schema[0].xpath(
            'xs:element[@name="whatever%s"]' % RESPONSE_SUFFIX, namespaces=ns)
        assert len(elts) > 0
        assert elts[0].attrib['type'] == 'xs:string'

    def test_bare_with_conflicting_types(self):
        class SomeService(Service):
            @srpc(Array(String), _returns=Array(String))
            def whatever(sa):
                return sa

            @srpc(Array(String), _returns=Array(String), _body_style='bare')
            def whatever_bare(sa):
                return sa

        app = Application([SomeService], tns='tns')
        app.transport = 'None'

        wsdl = Wsdl11(app.interface)
        wsdl.build_interface_document('url')
        wsdl = etree.fromstring(wsdl.get_interface_document())
        schema, = wsdl.xpath(
            '/wsdl:definitions/wsdl:types/xs:schema[@targetNamespace="tns"]',
            namespaces=ns,
        )

        print(etree.tostring(schema, pretty_print=True))

        assert len(schema.xpath(
            'xs:complexType[@name="string%s"]' % ARRAY_SUFFIX,
                                                             namespaces=ns)) > 0

        elts = schema.xpath(
            'xs:element[@name="whatever_bare%s"]' % REQUEST_SUFFIX,
                                                                  namespaces=ns)

        assert len(elts) > 0
        assert elts[0].attrib['type'] == 'tns:string%s' % ARRAY_SUFFIX

        elts = schema.xpath(
            'xs:element[@name="whatever_bare%s"]' % RESPONSE_SUFFIX,
                                                                  namespaces=ns)

        assert len(elts) > 0
        assert elts[0].attrib['type'] == 'tns:string%s' % ARRAY_SUFFIX


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

import logging
logging.basicConfig(level=logging.DEBUG)

from spyne import Application, rpc, ServiceBase, \
    Integer, Unicode

from spyne import Iterable

from spyne.protocol.soap import Soap11

from spyne.server.twisted import TwistedWebResource

from twisted.internet import reactor
from twisted.web.server import Site

class HelloWorldService(ServiceBase):
    @rpc(Unicode, Integer, _returns=Iterable(Unicode))
    def say_hello(ctx, name, times):
        for i in range(times):
            yield 'Hello, %s' % name

application = Application([HelloWorldService],
    tns='spyne.examples.hello',
    in_protocol=Soap11(validator='lxml'),
    out_protocol=Soap11()
)

if __name__ == '__main__':
    resource = TwistedWebResource(application)
    site = Site(resource)

    reactor.listenTCP(8000, site, interface='0.0.0.0')
    reactor.run()





#!/usr/bin/env python
# encoding: utf8
#
# Copyright © Burak Arslan <burak at arskom dot com dot tr>,
#             Arskom Ltd. http://www.arskom.com.tr
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#    1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#    3. Neither the name of the owner nor the names of its contributors may be
#       used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
# OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#


"""
This is a simple example that illustrates how to have immediate children of a
complexType in a different namespace.
"""

from spyne import Unicode, Iterable, XmlAttribute, ComplexModel, Service, \
    Application, rpc
from spyne.protocol.soap import Soap11
from spyne.server.wsgi import WsgiApplication


NS_B = "www.example.com/schema/b"


class Baz(ComplexModel):
    __namespace__ = NS_B

    Thing = Unicode
    AttrC = XmlAttribute(Unicode)


class FooCustomRequest(ComplexModel):
    AttrA = XmlAttribute(Unicode)
    AttrB = XmlAttribute(Unicode)
    Bar = Baz.customize(sub_ns=NS_B)
    Baz = Unicode


class FooService(Service):
    @rpc(FooCustomRequest, _returns = Iterable(Unicode), _body_style='bare')
    def Foo(ctx, req):
        AttrA, AttrB, Baz, Bar = req.AttrA, req.AttrB, req.Baz, req.Bar
        yield 'Hello, %s' % Bar


application = Application([FooService],
    tns="www.example.com/schema/a",
    in_protocol=Soap11(validator='soft'),
    out_protocol=Soap11(),
)


wsgi_application = WsgiApplication(application)


if __name__ == '__main__':
    import logging

    from wsgiref.simple_server import make_server

    logging.basicConfig(level=logging.DEBUG)
    logging.getLogger('spyne.protocol.xml').setLevel(logging.DEBUG)

    logging.info("listening to http://127.0.0.1:8000")
    logging.info("wsdl is at: http://localhost:8000/?wsdl")

    server = make_server('127.0.0.1', 8000, wsgi_application)
    server.serve_forever()




#
# spyne - Copyright (C) Spyne contributors.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
#


"""
A server that uses http as transport via wsgi. It doesn't contain any server
logic.
"""


import logging
logger = logging.getLogger(__name__)

import cgi
import threading

from inspect import isgenerator
from itertools import chain

from spyne import Address, File, Fault
from spyne.util.six.moves.http_cookies import SimpleCookie
from spyne.util.six.moves.urllib.parse import unquote, quote

from spyne.application import get_fault_string_from_exception
from spyne.auxproc import process_contexts
from spyne.error import RequestTooLongError
from spyne.protocol.http import HttpRpc
from spyne.server.http import HttpBase, HttpMethodContext, HttpTransportContext
from spyne.util.odict import odict
from spyne.util.address import address_parser

from spyne.const.ansi_color import LIGHT_GREEN
from spyne.const.ansi_color import END_COLOR
from spyne.const.http import HTTP_200
from spyne.const.http import HTTP_404
from spyne.const.http import HTTP_500


try:
    from spyne.protocol.soap.mime import apply_mtom
except ImportError as _import_error_1:
    _local_import_error_1 = _import_error_1  # python 3 workaround
    def apply_mtom(*args, **kwargs):
        raise _local_import_error_1


try:
    from werkzeug.formparser import parse_form_data
except ImportError as _import_error_2:
    _local_import_error_2 = _import_error_2  # python 3 workaround
    def parse_form_data(*args, **kwargs):
        raise _local_import_error_2


def _reconstruct_url(environ, protocol=True, server_name=True, path=True,
                                                             query_string=True):
    """Rebuilds the calling url from values found in the
    environment.

    This algorithm was found via PEP 333, the wsgi spec and
    contributed by Ian Bicking.
    """

    url = ''
    if protocol:
        url = environ['wsgi.url_scheme'] + '://'

    if server_name:
        if environ.get('HTTP_HOST'):
            url += environ['HTTP_HOST']

        else:
            url += environ['SERVER_NAME']

            if environ['wsgi.url_scheme'] == 'https':
                if environ['SERVER_PORT'] != '443':
                    url += ':' + environ['SERVER_PORT']

            else:
                if environ['SERVER_PORT'] != '80':
                    url += ':' + environ['SERVER_PORT']

    if path:
        if (quote(environ.get('SCRIPT_NAME', '')) == '/' and
            quote(environ.get('PATH_INFO', ''))[0] == '/'):
            #skip this if it is only a slash
            pass

        elif quote(environ.get('SCRIPT_NAME', ''))[0:2] == '//':
            url += quote(environ.get('SCRIPT_NAME', ''))[1:]

        else:
            url += quote(environ.get('SCRIPT_NAME', ''))

        url += quote(environ.get('PATH_INFO', ''))

    if query_string:
        if environ.get('QUERY_STRING'):
            url += '?' + environ['QUERY_STRING']

    return url


def _parse_qs(qs):
    pairs = (s2 for s1 in qs.split('&') for s2 in s1.split(';'))
    retval = odict()

    for name_value in pairs:
        if name_value is None or len(name_value) == 0:
            continue
        nv = name_value.split('=', 1)

        if len(nv) != 2:
            # Handle case of a control-name with no equal sign
            nv.append(None)

        name = unquote(nv[0].replace('+', ' '))

        value = None
        if nv[1] is not None:
            value = unquote(nv[1].replace('+', ' '))

        l = retval.get(name, None)
        if l is None:
            l = retval[name] = []
        l.append(value)

    return retval


def _get_http_headers(req_env):
    retval = {}

    for k, v in req_env.items():
        if k.startswith("HTTP_"):
            key = k[5:].lower()
            val = [v]
            retval[key]= val
            logger.debug("Add http header %r = %r", key, val)

    return retval


def _gen_http_headers(headers):
    retval = []

    for k,v in headers.items():
        if isinstance(v, (list, tuple)):
            for v2 in v:
                retval.append((k, v2))
        else:
            retval.append((k, v))

    return retval


class WsgiTransportContext(HttpTransportContext):
    """The class that is used in the transport attribute of the
    :class:`WsgiMethodContext` class."""

    def __init__(self, parent, transport, req_env, content_type):
        super(WsgiTransportContext, self).__init__(parent, transport,
                                                          req_env, content_type)

        self.req_env = self.req
        """WSGI Request environment"""

        self.req_method = req_env.get('REQUEST_METHOD', None)
        """HTTP Request verb, as a convenience to users."""

        self.headers = _get_http_headers(self.req_env)

    def get_url(self):
        return _reconstruct_url(self.req_env)

    def get_path(self):
        return self.req_env['PATH_INFO']

    def get_path_and_qs(self):
        retval = quote(self.req_env.get('PATH_INFO', ''))
        qs = self.req_env.get('QUERY_STRING', None)
        if qs is not None:
            retval += '?' + qs
        return retval

    def get_cookie(self, key):
        cookie_string = self.req_env.get('HTTP_COOKIE', None)
        if cookie_string is None:
            return

        cookie = SimpleCookie()
        cookie.load(cookie_string)

        return cookie.get(key, None).value

    def get_request_method(self):
        return self.req['REQUEST_METHOD'].upper()

    def get_request_content_type(self):
        return self.req.get("CONTENT_TYPE", None)

    def get_peer(self):
        addr, port = address_parser.get_ip(self.req),\
                                               address_parser.get_port(self.req)

        if address_parser.is_valid_ipv4(addr):
            return Address(type=Address.TCP4, host=addr, port=port)

        if address_parser.is_valid_ipv6(addr):
            return Address(type=Address.TCP6, host=addr, port=port)


class WsgiMethodContext(HttpMethodContext):
    """The WSGI-Specific method context. WSGI-Specific information is stored in
    the transport attribute using the :class:`WsgiTransportContext` class.
    """

    TransportContext = None
    HttpTransportContext = WsgiTransportContext


class WsgiApplication(HttpBase):
    """A `PEP-3333 <http://www.python.org/dev/peps/pep-3333>`_
    compliant callable class.

    If you want to have a hard-coded URL in the wsdl document, this is how to do
    it: ::

        wsgi_app = WsgiApplication(...)
        wsgi_app.doc.wsdl11.build_interface_document("http://example.com")

    This is not strictly necessary -- if you don't do this, Spyne will get the
    URL from the first request, build the wsdl on-the-fly and cache it as a
    string in memory for later requests. However, if you want to make sure
    you only have this url on the WSDL, this is how to do it. Note that if
    your client takes the information in the Wsdl document seriously (not all
    do), all requests will go to the designated url above even when you get the
    Wsdl from another location, which can make testing a bit difficult. Use in
    moderation.

    Supported events:
        * ``wsdl``
            Called right before the wsdl data is returned to the client.

        * ``wsdl_exception``
            Called right after an exception is thrown during wsdl generation.
            The exception object is stored in ctx.transport.wsdl_error
            attribute.

        * ``wsgi_call``
            Called first when the incoming http request is identified as a rpc
            request.

        * ``wsgi_return``
            Called right before the output stream is returned to the WSGI
            handler.

        * ``wsgi_exception``
            Called right before returning the exception to the client.

        * ``wsgi_close``
            Called after the whole data has been returned to the client. It's
            called both from success and error cases.
    """

    def __init__(self, app, chunked=True, max_content_length=2 * 1024 * 1024,
                                                         block_length=8 * 1024):
        super(WsgiApplication, self).__init__(app, chunked, max_content_length,
                                                                   block_length)

        self._mtx_build_interface_document = threading.Lock()

        self._wsdl = None
        if self.doc.wsdl11 is not None:
            self._wsdl = self.doc.wsdl11.get_interface_document()

    def __call__(self, req_env, start_response, wsgi_url=None):
        """This method conforms to the WSGI spec for callable wsgi applications
        (PEP 333). It looks in environ['wsgi.input'] for a fully formed rpc
        message envelope, will deserialize the request parameters and call the
        method on the object returned by the get_handler() method.
        """

        url = wsgi_url
        if url is None:
            url = _reconstruct_url(req_env).split('.wsdl')[0]

        if self.is_wsdl_request(req_env):
            # Format the url for location
            url = url.split('?')[0].split('.wsdl')[0]
            return self.handle_wsdl_request(req_env, start_response, url)

        else:
            return self.handle_rpc(req_env, start_response)

    def is_wsdl_request(self, req_env):
        # Get the wsdl for the service. Assume path_info matches pattern:
        # /stuff/stuff/stuff/serviceName.wsdl or
        # /stuff/stuff/stuff/serviceName/?wsdl

        return (
            req_env['REQUEST_METHOD'].upper() == 'GET'
            and (
                (
                    'QUERY_STRING' in req_env
                    and req_env['QUERY_STRING'].split('=')[0].lower() == 'wsdl'
                )
                or req_env['PATH_INFO'].endswith('.wsdl')
            )
        )

    def handle_wsdl_request(self, req_env, start_response, url):
        ctx = WsgiMethodContext(self, req_env, 'text/xml; charset=utf-8')

        if self.doc.wsdl11 is None:
            start_response(HTTP_404,
                                  _gen_http_headers(ctx.transport.resp_headers))
            return [HTTP_404]

        if self._wsdl is None:
            self._wsdl = self.doc.wsdl11.get_interface_document()

        ctx.transport.wsdl = self._wsdl

        if ctx.transport.wsdl is None:
            try:
                self._mtx_build_interface_document.acquire()

                ctx.transport.wsdl = self._wsdl

                if ctx.transport.wsdl is None:
                    self.doc.wsdl11.build_interface_document(url)
                    ctx.transport.wsdl = self._wsdl = \
                                        self.doc.wsdl11.get_interface_document()

            except Exception as e:
                logger.exception(e)
                ctx.transport.wsdl_error = e

                self.event_manager.fire_event('wsdl_exception', ctx)

                start_response(HTTP_500,
                                  _gen_http_headers(ctx.transport.resp_headers))

                return [HTTP_500]

            finally:
                self._mtx_build_interface_document.release()

        self.event_manager.fire_event('wsdl', ctx)

        ctx.transport.resp_headers['Content-Length'] = \
                                                    str(len(ctx.transport.wsdl))
        start_response(HTTP_200, _gen_http_headers(ctx.transport.resp_headers))

        retval = ctx.transport.wsdl

        ctx.close()

        return [retval]

    def handle_error(self, p_ctx, others, error, start_response):
        """Serialize errors to an iterable of strings and return them.

        :param p_ctx: Primary (non-aux) context.
        :param others: List if auxiliary contexts (can be empty).
        :param error: One of ctx.{in,out}_error.
        :param start_response: See the WSGI spec for more info.
        """

        if p_ctx.transport.resp_code is None:
            p_ctx.transport.resp_code = \
                p_ctx.out_protocol.fault_to_http_response_code(error)

        self.get_out_string(p_ctx)

        # consume the generator to get the length
        p_ctx.out_string = list(p_ctx.out_string)

        p_ctx.transport.resp_headers['Content-Length'] = \
                                    str(sum((len(s) for s in p_ctx.out_string)))
        self.event_manager.fire_event('wsgi_exception', p_ctx)

        start_response(p_ctx.transport.resp_code,
                                _gen_http_headers(p_ctx.transport.resp_headers))

        try:
            process_contexts(self, others, p_ctx, error=error)
        except Exception as e:
            # Report but ignore any exceptions from auxiliary methods.
            logger.exception(e)

        return chain(p_ctx.out_string, self.__finalize(p_ctx))

    def handle_rpc(self, req_env, start_response):
        initial_ctx = WsgiMethodContext(self, req_env,
                                                self.app.out_protocol.mime_type)

        self.event_manager.fire_event('wsgi_call', initial_ctx)
        initial_ctx.in_string, in_string_charset = \
                                        self.__reconstruct_wsgi_request(req_env)

        contexts = self.generate_contexts(initial_ctx, in_string_charset)
        p_ctx, others = contexts[0], contexts[1:]

        # TODO: rate limiting
        p_ctx.active = True

        if p_ctx.in_error:
            return self.handle_error(p_ctx, others, p_ctx.in_error,
                                                                 start_response)

        self.get_in_object(p_ctx)
        if p_ctx.in_error:
            logger.error(p_ctx.in_error)
            return self.handle_error(p_ctx, others, p_ctx.in_error,
                                                                 start_response)

        self.get_out_object(p_ctx)
        if p_ctx.out_error:
            return self.handle_error(p_ctx, others, p_ctx.out_error,
                                                                 start_response)

        assert p_ctx.out_object is not None
        g = next(iter(p_ctx.out_object))
        is_generator = len(p_ctx.out_object) == 1 and isgenerator(g)

        # if the out_object is a generator function, this hack makes the user
        # code run until first yield, which lets it set response headers and
        # whatnot before calling start_response. It's important to run this
        # here before serialization as the user function can also set output
        # protocol. Is there a better way?
        if is_generator:
            first_obj = next(g)
            p_ctx.out_object = ( chain((first_obj,), g), )

        if p_ctx.transport.resp_code is None:
            p_ctx.transport.resp_code = HTTP_200

        try:
            self.get_out_string(p_ctx)

        except Exception as e:
            logger.exception(e)
            p_ctx.out_error = Fault('Server', get_fault_string_from_exception(e))
            return self.handle_error(p_ctx, others, p_ctx.out_error,
                                                                 start_response)


        if isinstance(p_ctx.out_protocol, HttpRpc) and \
                                               p_ctx.out_header_doc is not None:
            p_ctx.transport.resp_headers.update(p_ctx.out_header_doc)

        if p_ctx.descriptor and p_ctx.descriptor.mtom:
            # when there is more than one return type, the result is
            # encapsulated inside a list. when there's just one, the result
            # is returned in a non-encapsulated form. the apply_mtom always
            # expects the objects to be inside an iterable, hence the
            # following test.
            out_type_info = p_ctx.descriptor.out_message._type_info
            if len(out_type_info) == 1:
                p_ctx.out_object = [p_ctx.out_object]

            p_ctx.transport.resp_headers, p_ctx.out_string = apply_mtom(
                    p_ctx.transport.resp_headers, p_ctx.out_string,
                    p_ctx.descriptor.out_message._type_info.values(),
                    p_ctx.out_object,
                )

        self.event_manager.fire_event('wsgi_return', p_ctx)

        if self.chunked:
            # the user has not set a content-length, so we delete it as the
            # input is just an iterable.
            if 'Content-Length' in p_ctx.transport.resp_headers:
                del p_ctx.transport.resp_headers['Content-Length']
        else:
            p_ctx.out_string = [''.join(p_ctx.out_string)]

        try:
            len(p_ctx.out_string)

            p_ctx.transport.resp_headers['Content-Length'] = \
                                    str(sum([len(a) for a in p_ctx.out_string]))
        except TypeError:
            pass

        start_response(p_ctx.transport.resp_code,
                                _gen_http_headers(p_ctx.transport.resp_headers))

        retval = chain(p_ctx.out_string, self.__finalize(p_ctx))

        try:
            process_contexts(self, others, p_ctx, error=None)
        except Exception as e:
            # Report but ignore any exceptions from auxiliary methods.
            logger.exception(e)

        return retval

    def __finalize(self, p_ctx):
        p_ctx.close()
        self.event_manager.fire_event('wsgi_close', p_ctx)

        return ()

    def __reconstruct_wsgi_request(self, http_env):
        """Reconstruct http payload using information in the http header."""

        content_type = http_env.get("CONTENT_TYPE")
        charset = None
        if content_type is not None:
            # fyi, here's what the parse_header function returns:
            # >>> import cgi; cgi.parse_header("text/xml; charset=utf-8")
            # ('text/xml', {'charset': 'utf-8'})
            content_type = cgi.parse_header(content_type)
            charset = content_type[1].get('charset', None)

        return self.__wsgi_input_to_iterable(http_env), charset

    def __wsgi_input_to_iterable(self, http_env):
        istream = http_env.get('wsgi.input')

        length = str(http_env.get('CONTENT_LENGTH', self.max_content_length))
        if len(length) == 0:
            length = 0
        else:
            length = int(length)

        if length > self.max_content_length:
            raise RequestTooLongError()
        bytes_read = 0

        while bytes_read < length:
            bytes_to_read = min(self.block_length, length - bytes_read)

            if bytes_to_read + bytes_read > self.max_content_length:
                raise RequestTooLongError()

            data = istream.read(bytes_to_read)
            if data is None or len(data) == 0:
                break

            bytes_read += len(data)

            yield data

    def decompose_incoming_envelope(self, prot, ctx, message):
        """This function is only called by the HttpRpc protocol to have the wsgi
        environment parsed into ``ctx.in_body_doc`` and ``ctx.in_header_doc``.
        """

        params = {}
        wsgi_env = ctx.in_document

        if self.has_patterns:
            # http://legacy.python.org/dev/peps/pep-0333/#url-reconstruction
            domain = wsgi_env.get('HTTP_HOST', None)
            if domain is None:
                domain = wsgi_env['SERVER_NAME']
            else:
                domain = domain.partition(':')[0] # strip port info

            params = self.match_pattern(ctx,
                    wsgi_env.get('REQUEST_METHOD', ''),
                    wsgi_env.get('PATH_INFO', ''),
                    domain,
                )

        if ctx.method_request_string is None:
            ctx.method_request_string = '{%s}%s' % (
                                    prot.app.interface.get_tns(),
                                    wsgi_env['PATH_INFO'].split('/')[-1])

        logger.debug("%sMethod name: %r%s" % (LIGHT_GREEN,
                                          ctx.method_request_string, END_COLOR))

        ctx.in_header_doc = ctx.transport.headers
        ctx.in_body_doc = _parse_qs(wsgi_env['QUERY_STRING'])

        for k, v in params.items():
             if k in ctx.in_body_doc:
                 ctx.in_body_doc[k].extend(v)
             else:
                 ctx.in_body_doc[k] = list(v)

        verb = wsgi_env['REQUEST_METHOD'].upper()
        if verb in ('POST', 'PUT', 'PATCH'):
            stream, form, files = parse_form_data(wsgi_env,
                                             stream_factory=prot.stream_factory)

            for k, v in form.lists():
                val = ctx.in_body_doc.get(k, [])
                val.extend(v)
                ctx.in_body_doc[k] = val

            for k, v in files.items():
                val = ctx.in_body_doc.get(k, [])

                mime_type = v.headers.get('Content-Type',
                                                     'application/octet-stream')

                path = getattr(v.stream, 'name', None)
                if path is None:
                    val.append(File.Value(name=v.filename, type=mime_type,
                                                    data=[v.stream.getvalue()]))
                else:
                    v.stream.seek(0)
                    val.append(File.Value(name=v.filename, type=mime_type,
                                                    path=path, handle=v.stream))

                ctx.in_body_doc[k] = val

            for k, v in ctx.in_body_doc.items():
                if v == ['']:
                    ctx.in_body_doc[k] = [None]




#
# spyne - Copyright (C) Spyne contributors.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
#

from twisted.internet.defer import Deferred
from twisted.internet.interfaces import IPullProducer
from twisted.web.iweb import UNKNOWN_LENGTH

from zope.interface import implementer


@implementer(IPullProducer)
class Producer(object):
    deferred = None

    def __init__(self, body, consumer):
        """:param body: an iterable of strings"""

        # check to see if we can determine the length
        try:
            len(body) # iterator?
            self.length = sum([len(fragment) for fragment in body])
            self.body = iter(body)

        except TypeError:
            self.length = UNKNOWN_LENGTH
            self.body = body

        self.deferred = Deferred()

        self.consumer = consumer

    def resumeProducing(self):
        try:
            chunk = next(self.body)

        except StopIteration as e:
            self.consumer.unregisterProducer()
            if self.deferred is not None:
                self.deferred.callback(self.consumer)
                self.deferred = None
            return

        self.consumer.write(chunk)

    def pauseProducing(self):
        pass

    def stopProducing(self):
        if self.deferred is not None:
            self.deferred.errback(
                               Exception("Consumer asked us to stop producing"))
        self.deferred = None


from spyne import Address
_TYPE_MAP = {'TCP': Address.TCP4, 'TCP6': Address.TCP6,
             'UDP': Address.UDP4, 'UDP6': Address.UDP6}

def _address_from_twisted_address(peer):
    return Address(
            type=_TYPE_MAP.get(peer.type, None), host=peer.host, port=peer.port)

Address.from_twisted_address = staticmethod(_address_from_twisted_address)



#!/usr/bin/env python

from __future__ import print_function

import json

from spyne import AnyUri, Unicode, ComplexModel, M, UnsignedInteger16, Array
from spyne.protocol.json import JsonDocument
from spyne.util.dictdoc import get_object_as_dict


class BenchmarkConfigElement(ComplexModel):
    # exclude this from the output document
    key = Unicode(pa={JsonDocument: dict(exc=True)})

    display_name = M(Unicode)
    notes = Unicode
    versus = Unicode

    db_url = AnyUri
    json_url = AnyUri
    query_url = AnyUri
    fortune_url = AnyUri
    update_url = AnyUri
    plaintext_url = AnyUri

    port = M(UnsignedInteger16(default=8080))

    approach = M(Unicode(values=['Realistic', 'Stripped'], default='Realistic'))
    classification = M(Unicode(values=['Micro', 'Fullstack', 'Platform'], default='Micro'))
    database = M(Unicode(values=['none', 'mongodb', 'postgres', 'mysql'], default='none'))
    orm = M(Unicode(values=['Full', 'Micro', 'None', 'Raw']))

    framework = M(Unicode)
    language = M(Unicode)
    flavor = M(Unicode)
    platform = M(Unicode)
    webserver = M(Unicode)

    os = M(Unicode(default='Linux'))
    database_os = M(Unicode(default='Linux'))
    
    
class BenchmarkConfig(ComplexModel):
    framework = M(Unicode)
    tests = Array(BenchmarkConfigElement, wrapped=False)


gen_raw_test = lambda: BenchmarkConfigElement(
    display_name="Spyne RAW",
    db_url="/dbsraw",
    query_url="/dbraw?queries=",
    fortune_url="/fortunesraw",
    update_url="/raw-updates?queries=",
    orm='Raw',
)

gen_normal_test = lambda: BenchmarkConfigElement(
    display_name="Spyne ORM",
    db_url="/dbs",
    query_url="/db?queries=",
    fortune_url="/fortunes",
    update_url="/updatesraw?queries=",
    orm='Full',
)


def add_common(bc):
    bc.port = 8080
    bc.approach = "Realistic"
    bc.classification = "Micro"
    bc.database = "postgres"
    bc.framework = "spyne"
    bc.language = "Python"
    bc.platform = "Spyne"
    bc.webserver = "None"
    bc.os = "Linux"
    bc.database_os = "Linux"
    bc.versus = "wsgi"
    bc.plaintext_url = "/plaintext"
    return bc


config = BenchmarkConfig(framework='spyne', tests=[])

keys = iter(['default', 'raw', 'py3orm', 'py3raw'])

for flav in ['CPython', 'Python3']:
    bc = add_common(gen_normal_test())
    bc.flavor = flav
    bc.key = next(keys)
    config.tests.append(bc)

    bc = add_common(gen_raw_test())
    bc.flavor = flav
    bc.key = next(keys)
    config.tests.append(bc)

data = get_object_as_dict(config, complex_as=dict)
data['tests'] = [{d['key']: d} for d in data['tests']]

data = json.dumps(data, indent=2, sort_keys=True, separators=(',', ': '))

open('benchmark_config.json', 'wb').write(data)


print(data)



# encoding: utf8
#
# spyne - Copyright (C) Spyne contributors.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
#

import logging
logger = logging.getLogger(__name__)

from collections import defaultdict

from lxml import etree, html
from lxml.html.builder import E

from spyne.util import coroutine, Break, six
from spyne.util.oset import oset
from spyne.util.etreeconv import dict_to_etree

from spyne.protocol.cloth import XmlCloth
from spyne.protocol.cloth._base import XmlClothProtocolContext


def parse_html_fragment_file(T_FILES):
    elt = html.fromstring(open(T_FILES).read())
    elt.getparent().remove(elt)
    return elt


class HtmlClothProtocolContext(XmlClothProtocolContext):
    def __init__(self, parent, transport, type=None):
        super(HtmlClothProtocolContext, self).__init__(parent, transport, type)

        self.assets = []
        self.eltstack = defaultdict(list)
        self.ctxstack = defaultdict(list)
        self.rootstack = oset()
        self.tags = set()
        self.objcache = dict()

        # these are supposed to be for neurons.base.screen.ScreenBase subclasses
        self.screen = None
        self.prev_view = None
        self.next_view = None


class HtmlCloth(XmlCloth):
    mime_type = 'text/html; charset=UTF-8'

    def __init__(self, app=None, encoding='utf8',
                      mime_type=None, ignore_uncap=False, ignore_wrappers=False,
                                cloth=None, cloth_parser=None, polymorphic=True,
                             strip_comments=True, hier_delim='.', doctype=None):

        super(HtmlCloth, self).__init__(app=app, encoding=encoding,
                                 mime_type=mime_type, ignore_uncap=ignore_uncap,
                                   ignore_wrappers=ignore_wrappers, cloth=cloth,
                             cloth_parser=cloth_parser, polymorphic=polymorphic,
                                                  strip_comments=strip_comments)

        self.hier_delim = hier_delim
        self.doctype = doctype
        self.default_method = 'html'

    def _parse_file(self, file_name, cloth_parser):
        if cloth_parser is None:
            cloth_parser = html.HTMLParser()

        cloth = html.parse(file_name, parser=cloth_parser)
        return cloth.getroot()

    def docfile(self, *args, **kwargs):
        logger.debug("Starting file with %r %r", args, kwargs)
        return etree.htmlfile(*args, **kwargs)

    def get_context(self, parent, transport):
        return HtmlClothProtocolContext(parent, transport)

    @staticmethod
    def get_class_cloth(cls):
        return cls.Attributes._html_cloth

    @staticmethod
    def get_class_root_cloth(cls):
        return cls.Attributes._html_root_cloth

    def dict_to_parent(self, ctx, cls, inst, parent, name, **kwargs):
        parent.write(repr(inst))

    @staticmethod
    def add_html_attr(attr_name, attr_dict, class_name):
        if attr_name in attr_dict:
            attr_dict[attr_name] = ' '.join(
                                       (attr_dict.get('class', ''), class_name))
        else:
            attr_dict[attr_name] = class_name

    @staticmethod
    def add_style(attr_dict, data):
        style = attr_dict.get('style', None)

        if style is not None:
            attr_dict['style'] = ';'.join(style, data)

        else:
            attr_dict['style'] = data

    @staticmethod
    def selsafe(s):
        return s.replace('[', '').replace(']', '').replace('.', '__')

    @coroutine
    def complex_to_parent(self, ctx, cls, inst, parent, name, use_ns=False,
                                                                      **kwargs):
        inst = cls.get_serialization_instance(inst)

        # TODO: Put xml attributes as well in the below element() call.
        with parent.element(name):
            ret = self._write_members(ctx, cls, inst, parent, use_ns=False,
                                                                       **kwargs)
            if ret is not None:
                try:
                    while True:
                        sv2 = (yield)  # may throw Break
                        ret.send(sv2)

                except Break:
                    try:
                        ret.throw(Break())
                    except StopIteration:
                        pass

    def gen_anchor(self, cls, inst, name, anchor_class=None):
        assert name is not None
        cls_attrs = self.get_cls_attrs(cls)

        href = getattr(inst, 'href', None)
        if href is None: # this is not a AnyUri.Value instance.
            href = inst

            content = None
            text = cls_attrs.text

        else:
            content = getattr(inst, 'content', None)
            text = getattr(inst, 'text', None)
            if text is None:
                text = cls_attrs.text

        if anchor_class is None:
            anchor_class = cls_attrs.anchor_class

        if text is None:
            text = name

        retval = E.a(text)

        if href is not None:
            retval.attrib['href'] = href

        if anchor_class is not None:
            retval.attrib['class'] = anchor_class

        if content is not None:
            retval.append(content)

        return retval

    def any_uri_to_parent(self, ctx, cls, inst, parent, name, **kwargs):
        retval = self.gen_anchor(cls, inst, name)
        parent.write(retval)

    def imageuri_to_parent(self, ctx, cls, inst, parent, name, **kwargs):
        # with ImageUri, content is ignored.
        href = getattr(inst, 'href', None)
        if href is None: # this is not a AnyUri.Value instance.
            href = inst
            text = getattr(cls.Attributes, 'text', None)

        else:
            text = getattr(inst, 'text', None)
            if text is None:
                text = getattr(cls.Attributes, 'text', None)

        retval = E.img(src=href)
        if text is not None:
            retval.attrib['alt'] = text

        parent.write(retval)

    def byte_array_to_parent(self, ctx, cls, inst, parent, name, **kwargs):
        ret = self.to_unicode(cls, inst, self.binary_encoding)

        if ret is not None:
            parent.write(ret)

    def model_base_to_parent(self, ctx, cls, inst, parent, name, **kwargs):
        ret = self.to_unicode(cls, inst)

        if ret is not None:
            parent.write(ret)

    def null_to_parent(self, ctx, cls, inst, parent, name, **kwargs):
        pass

    def any_xml_to_parent(self, ctx, cls, inst, parent, name, **kwargs):
        if isinstance(inst, (six.text_type, six.binary_type)):
            inst = etree.fromstring(inst)

        parent.write(inst)

    def any_html_to_parent(self, ctx, cls, inst, parent, name, **kwargs):
        cls_attrs = self.get_cls_attrs(cls)

        if cls_attrs.as_string:
            if not (isinstance(inst, str) or isinstance(inst, six.text_type)):
                inst = html.tostring(inst)

        else:
            if isinstance(inst, str) or isinstance(inst, six.text_type):
                inst = html.fromstring(inst)

        parent.write(inst)

    def any_to_parent(self, ctx, cls, inst, parent, name, **kwargs):
        parent.write(inst)

    def any_dict_to_parent(self, ctx, cls, inst, parent, name, **kwargs):
        elt = E('foo')
        dict_to_etree(inst, elt)

        parent.write(elt[0])

    def fault_to_parent(self, ctx, cls, inst, parent, name, **kwargs):
        self.complex_to_parent(ctx, cls, inst, parent, name, **kwargs)


# FIXME: Deprecated
HtmlBase = HtmlCloth



#!/usr/bin/env python
# encoding: utf8
#
# Copyright   Burak Arslan <burak at arskom dot com dot tr>,
#             Arskom Ltd. http://www.arskom.com.tr
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#    1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#    3. Neither the name of the owner nor the names of its contributors may be
#       used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
# OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import logging

from pyramid.config import Configurator
from pyramid.view import view_config

from wsgiref.simple_server import make_server

from spyne.util.simple import pyramid_soap11_application
from spyne.decorator import rpc
from spyne.model.complex import Iterable
from spyne.model.primitive import Integer
from spyne.model.primitive import String
from spyne.service import Service

logging.basicConfig()
logging.basicConfig(level=logging.DEBUG)
logging.getLogger('spyne.protocol.xml').setLevel(logging.DEBUG)

tns = 'spyne.examples.pyramid.helloworld'

'''
>>> from suds.client import Client
>>> cli = Client("http://localhost:8000/?wsdl")
>>> cli
<suds.client.Client object at 0x104d5ed10>
>>> print cli

Suds ( https://fedorahosted.org/suds/ )  version: 0.4 GA  build: R699-20100913

Service ( HelloWorldService ) tns="spyne.helloworld"
   Prefixes (1)
      ns0 = "spyne.helloworld"
   Ports (1):
      (Application)
         Methods (1):
            say_hello(xs:string name, xs:integer times, )
         Types (3):
            say_hello
            say_helloResponse
            stringArray


>>> res = cli.service.say_hello('justin', 5)
>>> res
(stringArray){
   string[] = 
      "Hello, justin",
      "Hello, justin",
      "Hello, justin",
      "Hello, justin",
      "Hello, justin",
 }
'''

class HelloWorldService(Service):
    __namespace__ = tns
    
    @rpc(String, Integer, _returns=Iterable(String))
    def say_hello(ctx, name, times):
        '''
        Docstrings for service methods appear as documentation in the wsdl
        <b>what fun</b>
        @param name the name to say hello to
        @param the number of times to say hello
        @return the completed array
        '''

        for i in range(times):
            yield 'Hello, %s' % name

# view
soapApp = view_config(route_name="home")(
    pyramid_soap11_application([HelloWorldService], tns))

if __name__ == '__main__':
    # configuration settings
    settings = {}
    settings['debug_all'] = True
    # configuration setup
    config = Configurator(settings=settings)
    # routes setup
    config.add_route('home', '/')
    config.scan()
    # serve app
    app = config.make_wsgi_app()
    server = make_server('0.0.0.0', 8000, app)
    server.serve_forever()



#!/usr/bin/env python

from __future__ import print_function

import sys

from lxml import etree
from spyne.util import six

from spyne import ComplexModel, Unicode
from spyne.util.xml import get_object_as_xml_polymorphic


class B(ComplexModel):
    _type_info = [
        ('_b', Unicode(default="b")),
    ]


class C(B):
    _type_info = [
        ('_c', Unicode(default="c")),
    ]


class A(ComplexModel):
    _type_info = [
        ('a', Unicode(subname="_a")),
        ('b', B.customize(subname="_b")),
    ]


a = A(b=C())
elt = get_object_as_xml_polymorphic(a, A, no_namespace=True)
xml_string = etree.tostring(elt, pretty_print=True)
if six.PY2:
    print(xml_string, end="")
else:
    sys.stdout.buffer.write(xml_string)
